"""Test configuration"""

import itertools
import sys
import uuid
from datetime import datetime, timezone
from pathlib import Path
from typing import Generator
from unittest.mock import Mock

import pytest
from dateutil import tz

from icalendar import (
    Alarm,
    Availability,
    Calendar,
    Component,
    ComponentFactory,
    Event,
    Timezone,
    TypesFactory,
    prop,
    vUTCOffset,
)
from icalendar.compatibility import ZoneInfo, zoneinfo
from icalendar.tests.data import PROPERTY_NAMES
from icalendar.timezone import TZP
from icalendar.timezone import tzp as _tzp

from . import timezone_ids

try:
    import pytz
except ImportError:
    pytz = None

HAS_PYTZ = pytz is not None
if HAS_PYTZ:
    PYTZ_UTC = [
        pytz.utc,
        pytz.timezone("UTC"),
    ]
    PYTZ_IN_TIMEZONE = [
        lambda dt, tzname: pytz.timezone(tzname).localize(dt),
    ]
    PYTZ_TZP = ["pytz"]
else:
    PYTZ_UTC = []
    PYTZ_IN_TIMEZONE = []
    PYTZ_TZP = []


class DataSource:
    """A collection of parsed ICS elements (e.g calendars, timezones, events)"""

    extensions = [".ics", ".jcal"]

    def __init__(self, data_source_folder: Path, parser):
        self._parser = parser
        self._data_source_folder = data_source_folder

    def keys(self):
        """Return all the files that could be used."""
        return [
            p.stem
            for p in self._data_source_folder.iterdir()
            if p.suffix.lower() in self.extensions
        ]

    def __getitem__(self, attribute):
        """Parse a file and return the result stored in the attribute."""
        extensions = self.extensions
        if "." in attribute:
            # we have a file ending
            attribute, extension = attribute.rsplit(".", 1)
            extensions = ["." + extension]
        for extension in extensions:
            source_file = attribute + extension
            source_path = self._data_source_folder / source_file
            if source_path.is_file():
                break
        if not source_path.is_file():
            # usually only raised if the test is wrong
            raise AttributeError(
                f"{attribute} does not exist with these extensions: {', '.join(extensions)}."
            )
        if extension == ".jcal":
            raw_jcal = source_path.read_text()
            source = Component.from_jcal(raw_jcal)
            raw_ics = None
        else:
            raw_ics = source_path.read_bytes()
            source = self._parser(raw_ics)
            raw_jcal = None
        if not isinstance(source, list):
            source.raw_ics = raw_ics
            source.raw_jcal = raw_jcal
            source.source_file = source_file
        self.__dict__[attribute] = source
        return source

    def __contains__(self, key):
        """key in self.keys()"""
        key = key.rsplit(".", 1)[0]
        return key in self.keys()

    def __getattr__(self, key):
        return self[key]

    def __repr__(self):
        return repr(self.__dict__)

    @property
    def multiple(self):
        """Return a list of all components parsed."""
        return self.__class__(
            self._data_source_folder, lambda data: self._parser(data, multiple=True)
        )


HERE = Path(__file__).parent
CALENDARS_FOLDER = HERE / "calendars"
TIMEZONES_FOLDER = HERE / "timezones"
EVENTS_FOLDER = HERE / "events"
ALARMS_FOLDER = HERE / "alarms"
AVAILABILITIES_FOLDER = HERE / "availabilities"


@pytest.fixture(scope="module")
def calendars(tzp):
    return DataSource(CALENDARS_FOLDER, Calendar.from_ical)


@pytest.fixture(scope="module")
def timezones(tzp):
    return DataSource(TIMEZONES_FOLDER, Timezone.from_ical)


@pytest.fixture(scope="module")
def events(tzp):
    return DataSource(EVENTS_FOLDER, Event.from_ical)


@pytest.fixture(scope="module")
def alarms(tzp):
    return DataSource(ALARMS_FOLDER, Alarm.from_ical)


@pytest.fixture(scope="module")
def availabilities(tzp):
    return DataSource(AVAILABILITIES_FOLDER, Availability.from_ical)


@pytest.fixture(
    params=PYTZ_UTC + [ZoneInfo("UTC"), tz.UTC, tz.gettz("UTC"), timezone.utc]
)
def utc(request, tzp):
    return request.param


@pytest.fixture(
    params=PYTZ_IN_TIMEZONE
    + [
        lambda dt, tzname: dt.replace(tzinfo=tz.gettz(tzname)),
        lambda dt, tzname: dt.replace(tzinfo=ZoneInfo(tzname)),
    ]
)
def in_timezone(request, tzp):
    return request.param


FUZZ_TESTCASES_BROKEN_CALENDARS = "fuzz_testcase"

# exclude broken calendars here
BROKEN_SOURCE_FILES = (
    "big_bad_calendar.ics",
    "issue_104_broken_calendar.ics",
    "small_bad_calendar.ics",
    "multiple_calendar_components.ics",
    "pr_480_summary_with_colon.ics",
    "parsing_error_in_UTC_offset.ics",
    "parsing_error.ics",
)
SOURCE_FILES = [
    file.name
    for file in itertools.chain(
        CALENDARS_FOLDER.iterdir(), TIMEZONES_FOLDER.iterdir(), EVENTS_FOLDER.iterdir()
    )
    if file.name not in BROKEN_SOURCE_FILES
    and file.suffix in (".ics", ".jcal")
    and FUZZ_TESTCASES_BROKEN_CALENDARS not in file.name
]

ICS_FILES = [
    file.name
    for file in itertools.chain(
        CALENDARS_FOLDER.iterdir(), TIMEZONES_FOLDER.iterdir(), EVENTS_FOLDER.iterdir()
    )
    if file.name not in BROKEN_SOURCE_FILES
    and file.suffix in (".ics",)
    and FUZZ_TESTCASES_BROKEN_CALENDARS not in file.name
]

JCAL_FILES = [
    file.name
    for file in itertools.chain(
        CALENDARS_FOLDER.iterdir(), TIMEZONES_FOLDER.iterdir(), EVENTS_FOLDER.iterdir()
    )
    if file.name not in BROKEN_SOURCE_FILES
    and file.suffix in (".jcal",)
    and FUZZ_TESTCASES_BROKEN_CALENDARS not in file.name
]


def get_source_file(calendars, timezones, events, request) -> Component:
    source_file = request.param
    print("example file:", source_file)
    for data in calendars, timezones, events:
        if source_file in data:
            return data[source_file]
    raise ValueError(f"Could not find file {source_file}.")


@pytest.fixture(params=SOURCE_FILES)
def source_file(tzp, calendars, timezones, events, request) -> Component:
    """An example file.

    source_file.raw_ics - bytes if .ics file
    source_file.raw_jcal - str if .jcal file

    """
    return get_source_file(calendars, timezones, events, request)


@pytest.fixture(params=ICS_FILES)
def ics_file(tzp, calendars, timezones, events, request) -> Component:
    """An example .ica file.

    source_file.raw_ics - bytes
    source_file.raw_jcal - None
    """
    return get_source_file(calendars, timezones, events, request)


@pytest.fixture(params=JCAL_FILES)
def jcal_file(tzp, calendars, timezones, events, request) -> Component:
    """An example .jcal file.

    source_file.raw_ics - None
    source_file.raw_jcal - str
    """
    return get_source_file(calendars, timezones, events, request)


FUZZ_V1 = [
    key
    for key in CALENDARS_FOLDER.iterdir()
    if FUZZ_TESTCASES_BROKEN_CALENDARS in str(key)
]


@pytest.fixture(params=FUZZ_V1)
def fuzz_v1_calendar_path(request):
    """Clusterfuzz calendars."""
    return request.param


@pytest.fixture
def types_factory():
    """Return a new types factory."""
    return TypesFactory()


@pytest.fixture
def component_factory():
    """Return a new types factory."""
    return ComponentFactory()


@pytest.fixture
def x_sometime(types_factory):
    """Map x_sometime to time"""
    types_factory.types_map["X-SOMETIME"] = "time"
    yield
    types_factory.types_map.pop("X-SOMETIME")


@pytest.fixture
def factory():
    """Return a new component factory."""
    return ComponentFactory()


@pytest.fixture
def vUTCOffset_ignore_exceptions():
    vUTCOffset.ignore_exceptions = True
    yield
    vUTCOffset.ignore_exceptions = False


@pytest.fixture
def event_component(tzp):
    """Return an event component."""
    c = Component()
    c.name = "VEVENT"
    return c


@pytest.fixture
def c(tzp):
    """Return an empty component."""
    return Component()


comp = c


@pytest.fixture
def calendar_component(tzp):
    """Return an empty component."""
    c = Component()
    c.name = "VCALENDAR"
    return c


@pytest.fixture
def dont_validate_new():
    """Remove validation for new() of components."""
    value = Component._validate_new
    Component._validate_new = False
    yield
    Component._validate_new = value


@pytest.fixture
def filled_event_component(c, calendar_component):
    """Return an event with some values and add it to calendar_component."""
    e = Component(summary="A brief history of time")
    e.name = "VEVENT"
    e.add("dtend", "20000102T000000", encode=0)
    e.add("dtstart", "20000101T000000", encode=0)
    calendar_component.add_component(e)
    return e


@pytest.fixture
def calendar_with_resources(tzp):
    c = Calendar()
    c["resources"] = 'Chair, Table, "Room: 42"'
    return c


@pytest.fixture(scope="module")
def tzp(tzp_name) -> Generator[TZP, None, None]:
    """The timezone provider."""
    _tzp.use(tzp_name)
    yield _tzp
    _tzp.use_default()


@pytest.fixture(params=PYTZ_TZP + ["zoneinfo"])
def other_tzp(request, tzp):
    """This is annother timezone provider.

    The purpose here is to cross test: pytz <-> zoneinfo.
    tzp as parameter makes sure we test the cross product.
    """
    return TZP(request.param)


@pytest.fixture
def pytz_only(tzp, tzp_name) -> str:
    """Skip tests that are not running under pytz."""
    assert tzp.uses_pytz()
    return tzp_name


@pytest.fixture
def zoneinfo_only(tzp, request, tzp_name) -> str:
    """Skip tests that are not running under zoneinfo."""
    assert tzp.uses_zoneinfo()
    return tzp_name


@pytest.fixture
def no_pytz(tzp_name) -> str:
    """Do not run tests with pytz."""
    assert tzp_name != "pytz"
    return tzp_name


@pytest.fixture
def no_zoneinfo(tzp_name) -> str:
    """Do not run tests with zoneinfo."""
    assert tzp_name != "zoneinfo"
    return tzp_name


def pytest_generate_tests(metafunc):
    """Parametrize without skipping:

    tzp_name will be parametrized according to the use of
    - pytz_only
    - zoneinfo_only
    - no_pytz
    - no_zoneinfo

    See https://docs.pytest.org/en/6.2.x/example/parametrize.html#deferring-the-setup-of-parametrized-resources
    """
    if "tzp_name" in metafunc.fixturenames:
        tzp_names = PYTZ_TZP + ["zoneinfo"]
        if "zoneinfo_only" in metafunc.fixturenames:
            tzp_names = ["zoneinfo"]
        if "pytz_only" in metafunc.fixturenames:
            tzp_names = PYTZ_TZP
        assert not (
            "zoneinfo_only" in metafunc.fixturenames
            and "pytz_only" in metafunc.fixturenames
        ), "Use pytz_only or zoneinfo_only but not both!"
        for name in ["pytz", "zoneinfo"]:
            if f"no_{name}" in metafunc.fixturenames and name in tzp_names:
                tzp_names.remove(name)
        metafunc.parametrize("tzp_name", tzp_names, scope="module")


class DoctestZoneInfo(ZoneInfo):
    """Constent ZoneInfo representation for tests."""

    def __repr__(self):
        return f"ZoneInfo(key={self.key!r})"


def doctest_print(obj):
    """doctest print"""
    if isinstance(obj, bytes):
        obj = obj.decode("UTF-8")
    print(str(obj).strip().replace("\r\n", "\n").replace("\r", "\n"))


def doctest_import(name, *args, **kw):
    """Replace the import mechanism to skip the whole doctest if we import pytz."""
    if name == "pytz":
        return pytz
    return __import__(name, *args, **kw)


NOW = datetime(2025, 5, 17, 8, 6, 12)
NOW_UTC = NOW.replace(tzinfo=timezone.utc)
UID_DEFAULT = "d755cef5-2311-46ed-a0e1-6733c9e15c63"


@pytest.fixture
def test_uid() -> str:
    """The UID that tests always create."""
    return UID_DEFAULT


@pytest.fixture(autouse=True)
def fixed_env(monkeypatch):
    """Create a fixed test environment for the functions that are time and randomness dependent."""
    # uuid is fixed for all tests
    uid = uuid.UUID(UID_DEFAULT, version=4)
    monkeypatch.setattr(uuid, "uuid4", lambda: uid)
    # now is fixed
    monkeypatch.setattr(Component, "_utc_now", staticmethod(lambda: NOW_UTC))


@pytest.fixture
def env_for_doctest(monkeypatch):
    """Modify the environment to make doctests run."""
    monkeypatch.setitem(sys.modules, "zoneinfo", zoneinfo)
    monkeypatch.setattr(zoneinfo, "ZoneInfo", DoctestZoneInfo)
    from icalendar.timezone.zoneinfo import ZONEINFO

    monkeypatch.setattr(ZONEINFO, "utc", zoneinfo.ZoneInfo("UTC"))
    return {"print": doctest_print}


@pytest.fixture(params=timezone_ids.TZIDS)
def tzid(request: pytest.FixtureRequest) -> str:
    """Return a timezone id to be used with pytz or zoneinfo.

    This goes through all the different timezones possible.
    """
    return request.param


@pytest.fixture(params=PROPERTY_NAMES)
def v_prop_name(request):
    """Names of property types that occur as a property of a component."""
    return request.param


@pytest.fixture
def v_prop(v_prop_name):
    """Property types that occur as a property of a component."""
    return getattr(prop, v_prop_name)


@pytest.fixture
def v_prop_example(v_prop) -> prop.VPROPERTY:
    return v_prop.examples()[0]


@pytest.fixture
def mock():
    """A mock."""
    return Mock()



"""Shared data between tests."""

from icalendar import prop

PROPERTY_NAMES = {attr for attr in dir(prop) if attr[0] == "v" and attr[1].isupper()}
# remove values for now that are in parameters only
PARAMETER_NAMES = {"vSkip", "vWeekday", "vFrequency", "vMonth", "vInline"}
PROPERTY_NAMES -= PARAMETER_NAMES


__all__ = ["PARAMETER_NAMES", "PROPERTY_NAMES"]



import datetime
import os

from icalendar import Calendar


def test_value_type_is_not_mapped(types_factory):
    """Usually, the value should be absent."""
    assert "X-SOMETIME" not in types_factory.types_map


def test_value_type_is_mapped(x_sometime, types_factory):
    """The value is mapped for the test."""
    assert "X-SOMETIME" in types_factory.types_map


def test_create_from_ical(x_sometime):
    directory = os.path.dirname(__file__)  # noqa: PTH120
    ics = open(os.path.join(directory, "calendars", "time.ics"), "rb")  # noqa: PTH118, PTH123, SIM115
    cal = Calendar.from_ical(ics.read())
    ics.close()

    assert cal["X-SOMETIME"].dt == datetime.time(17, 20, 10)
    assert cal["X-SOMETIME"].to_ical() == "172010"


def test_create_to_ical(x_sometime):
    cal = Calendar()
    cal.add("X-SOMETIME", datetime.time(17, 20, 10))
    assert b"X-SOMETIME;VALUE=TIME:172010" in cal.to_ical().splitlines()



from icalendar.prop import (
    vBoolean,
    vInline,
    vUTCOffset,
    vCategory,
    vCalAddress,
    vWeekday,
    vDuration,
    vFloat,
    vGeo,
    vInt,
    vText,
    vMonth,
    vUTCOffset,
    vFrequency,
    vRecur,
    vDatetime,
    vUri,
)
import datetime


def test_param_vCategory():
    obj = vCategory(["Work", "Personal"], params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vCategory)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vCalAddress():
    obj = vCalAddress("mailto:jane_doe@example.com", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vCalAddress)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vWeekday():
    obj = vWeekday("2FR", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vWeekday)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vBoolean():
    obj = vBoolean(True, params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vBoolean)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vDuration():
    td = datetime.timedelta(days=15, seconds=18020)
    obj = vDuration(td, params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vDuration)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vFloat():
    obj = vFloat("1.333", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vFloat)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vGeo():
    obj = vGeo((37.386013, -122.082932), params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vGeo)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vInt():
    obj = vInt("87", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vInt)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vInline():
    obj = vInline("sometxt", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vInline)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vText():
    obj = vText("sometxt", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vText)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vMonth():
    obj = vMonth(1, params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vMonth)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vUTCOffset():
    obj = vUTCOffset(
        datetime.timedelta(days=-1, seconds=68400), params={"SOME_PARAM": "VALUE"}
    )
    assert isinstance(obj, vUTCOffset)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vFrequency():
    obj = vFrequency("DAILY", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vFrequency)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vRecur():
    obj = vRecur({"FREQ": ["DAILY"], "COUNT": [10]}, params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vRecur)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vDatetime():
    dt = datetime.datetime(2025, 3, 16, 14, 30, 0, tzinfo=datetime.timezone.utc)
    obj = vDatetime(dt, params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vDatetime)
    assert obj.params["SOME_PARAM"] == "VALUE"


def test_param_vUri():
    obj = vUri("WWW.WESBITE.COM", params={"SOME_PARAM": "VALUE"})
    assert isinstance(obj, vUri)
    assert obj.params["SOME_PARAM"] == "VALUE"



"""Events differ although their times are equal.

See https://github.com/collective/icalendar/issues/828
"""

import contextlib
from datetime import date, datetime, timedelta, timezone

import pytest

from icalendar import Event, Journal, vDate, vDatetime, vDDDLists, vDDDTypes
from icalendar.compatibility import ZoneInfo


def to_dt(a: date) -> date:
    return a


def to_event(a):
    e = Event()
    e.start = a
    e.end = a + timedelta(days=1)
    e.add("RECURRENCE-ID", a - timedelta(days=1))
    return e


def to_journal(a):
    """return a journal for testing"""
    j = Journal()
    j.start = a
    j.end = a + timedelta(days=1)
    j.add("RECURRENCE-ID", a - timedelta(days=1))
    return j


def to_vDD(a):
    """Return a value type."""
    if isinstance(a, datetime):
        return vDatetime(a)
    return vDate(a)


def to_vDDDTypes(a):
    return vDDDTypes(a)


def to_vDDDLists(a):
    return vDDDLists([a])


equal_dt_pairs = [
    (date(1998, 10, 1), date(1998, 10, 1)),
    (datetime(2023, 12, 31), datetime(2023, 12, 31)),
    (
        datetime(2023, 12, 31, tzinfo=timezone.utc),
        datetime(2023, 12, 31, tzinfo=ZoneInfo("UTC")),
    ),
    (
        datetime(2023, 12, 31, tzinfo=ZoneInfo("UTC")),
        datetime(2023, 12, 31, tzinfo=ZoneInfo("UTC")),
    ),
    (
        datetime(2025, 12, 31, tzinfo=ZoneInfo("UTC")),
        datetime(2025, 12, 31, tzinfo=ZoneInfo("GMT+0")),
    ),
    (
        datetime(2025, 12, 31, 12, tzinfo=ZoneInfo("Europe/Zurich")),
        datetime(2025, 12, 31, 11, tzinfo=ZoneInfo("UTC")),
    ),
]


param_equal_dts = pytest.mark.parametrize(("d1", "d2"), equal_dt_pairs)

param_transform = pytest.mark.parametrize(
    "transform", [to_dt, to_event, to_journal, to_vDD, to_vDDDTypes, to_vDDDLists]
)


@param_equal_dts
@param_transform
def test_equality_of_equal_datetimes(d1, d2, transform):
    """Check that the values are equal."""
    a = transform(d1)
    b = transform(d2)
    assert_equal(a, b)


def assert_equal(a, b):
    """Check all equality tests."""
    assert a == b, f"1 equal: {a} == {b}"
    assert b == a, f"2 equal reversed: {b} == {a}"
    assert not a != b, f"3 not equal: {a} != {b}"  # noqa: SIM202
    assert not b != a, f"4 not equal reversed: {b} != {a}"  # noqa: SIM202
    assert bool(a) == bool(b), f"5 equal bool: {bool(a)} == {bool(b)}"
    with contextlib.suppress(TypeError):
        assert hash(a) == hash(b)


def assert_not_eq(a, b):
    """Check all equality tests."""
    assert a != b, f"1 unequal: {a} == {b}"
    assert b != a, f"2 unequal reversed: {b} == {a}"
    assert not a == b, f"3 not unequal: {a} != {b}"  # noqa: SIM201
    assert not b == a, f"4 not unequal reversed: {b} != {a}"  # noqa: SIM201


dts = [to_dt, to_vDD, to_vDDDTypes, to_vDDDLists]


@pytest.mark.parametrize("transform1", dts)
@pytest.mark.parametrize("transform2", dts)
@param_equal_dts
def test_property_times_for_date_and_datetime(d1, d2, transform1, transform2):
    """Check the equality of the date and date implementations."""
    a = transform1(d1)
    b = transform2(d2)
    assert_equal(a, b)


def test_datetime_is_in_list_representation():
    dt = datetime(2023, 12, 31)
    assert str(dt) in str(vDDDLists([dt]))


unequal_pairs = [(a, b) for a, b in equal_dt_pairs if a != b]


@pytest.mark.parametrize(("d1", "d2"), unequal_pairs)
@param_transform
def test_inequality(d1, d2, transform):
    """The values are not equal."""
    a = transform(d1)
    b = transform(d2)
    assert_not_eq(a, b)


@pytest.mark.parametrize("ddd_type", [to_vDD, to_vDDDTypes])
@param_equal_dts
def test_not_equal_if_parameters_differ(d1, d2, ddd_type):
    """If the items are equal but the parameters differ, they should not be equal."""
    d1 = ddd_type(d1)
    d2 = ddd_type(d2)
    d1.params["foo"] = "bar"
    assert_not_eq(d1, d2)


@pytest.mark.parametrize("ddd_type", [to_vDD, to_vDDDTypes])
@param_equal_dts
def test_equal_ignoring_x_params(d1, d2, ddd_type):
    """RFC 5545: Applications MUST ignore x-param and iana-param values they don't recognize."""
    d1 = ddd_type(d1)
    d2 = ddd_type(d2)
    d1.params["x-foo"] = "bar"
    assert_equal(d1, d2)


def test_list_is_not_hashable():
    """We cannot hash a list because it changes."""
    with pytest.raises(TypeError):
        hash(vDDDLists([datetime(2023, 12, 31)]))



"""We want to add rrule as a string for convenience.

See https://github.com/collective/icalendar/issues/301
"""

from icalendar.cal.event import Event


def test_rrule_add_example():
    event = Event()
    event.add("RRULE", "FREQ=DAILY;INTERVAL=2")
    assert "FREQ=DAILY;INTERVAL=2" in event.to_ical().decode("utf-8")
    assert event.rrules[0]["freq"] == ["DAILY"]
    assert event.rrules[0]["interval"] == [2]



"""This test tests all fuzzed calendars.

To add a new calendar test, echo the base64 calendar string into a file.

- The file is in src/icalendar/tests/calendars
- The file name contains "fuzz_testcase" and ends with ".ics"

Templace code to create a new test case:

.. code-block:: console

    echo "" | base64 -d > src/icalendar/tests/calendars/fuzz_testcase_.ics

"""

import icalendar.cal.calendar
from icalendar.tests.fuzzed import fuzz_v1_calendar


def test_fuzz_v1(fuzz_v1_calendar_path):
    """Test a calendar."""
    calendar = fuzz_v1_calendar_path.read_bytes()
    print(repr(calendar))
    fuzz_v1_calendar(
        icalendar.cal.calendar.Calendar.from_ical,
        calendar,
        multiple=True,
        should_walk=True,
    )



"""Test that we can identify all timezones.

Timezones can be removed from ./timezone_ids.py if they make the tests fail:
Timezone information changes over time and can be dependent on the operating system's
timezone database (zoneinfo, dateutil) or the package (pytz).
We want to make sure we can roughly identify most of them.
"""

from datetime import datetime, timedelta, timezone

import pytest

from icalendar.timezone import is_utc, tzid_from_tzinfo, tzids_from_tzinfo
from icalendar.timezone.tzp import TZP


def test_can_identify_zoneinfo(tzid, zoneinfo_only, tzp: TZP):
    """Check that all those zoneinfo timezones can be identified."""
    assert tzid in tzids_from_tzinfo(tzp.timezone(tzid))


def test_can_identify_pytz(tzid, pytz_only, tzp: TZP):
    """Check that all those pytz timezones can be identified."""
    assert tzid in tzids_from_tzinfo(tzp.timezone(tzid))


def test_can_identify_dateutil(tzid):
    """Check that all those dateutil timezones can be identified."""
    from dateutil.tz import gettz

    assert tzid in tzids_from_tzinfo(gettz(tzid))


def test_utc_is_identified(utc):
    """Test UTC because it is handled in a special way."""
    assert "UTC" in tzids_from_tzinfo(utc)
    assert tzid_from_tzinfo(utc) == "UTC"
    assert is_utc(utc)


def test_some_timezones_are_not_utc(tzp: TZP):
    """If we have an offset, it is not utc."""
    assert not is_utc(tzp.timezone("Europe/Berlin"))


def test_some_timezones_are_recognized_as_utc():
    """If we are at 0 offset, it is utc."""
    tzinfo = timezone(timedelta(hours=0))
    assert is_utc(tzinfo)
    assert is_utc(datetime(2019, 1, 1, 0, 0, 0, 0, tzinfo=tzinfo))


@pytest.mark.parametrize(
    "tzinfo",
    [
        timezone(timedelta(hours=1)),
        timezone(timedelta(hours=-1)),
        timezone(timedelta(hours=2)),
        timezone(timedelta(hours=-2)),
    ],
)
def test_getting_a_timezone_name_for_timezones(tzinfo):
    """We should get a name that we can use."""
    assert tzid_from_tzinfo(tzinfo)



"""This module identifies timezones.

Normally, timezones have ids.
This is a way to access the ids if you have a
datetime.tzinfo object.
"""

from __future__ import annotations

from collections import defaultdict
from datetime import date, time, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from dateutil.tz import tz

from icalendar.timezone import equivalent_timezone_ids_result
from icalendar.tools import is_date

if TYPE_CHECKING:
    from datetime import datetime, tzinfo

DATEUTIL_UTC = tz.gettz("UTC")
DATEUTIL_UTC_PATH: Optional[str] = getattr(DATEUTIL_UTC, "_filename", None)
DATEUTIL_ZONEINFO_PATH = (
    None if DATEUTIL_UTC_PATH is None else Path(DATEUTIL_UTC_PATH).parent
)


def tzids_from_tzinfo(tzinfo: Optional[tzinfo]) -> tuple[str]:
    """Get several timezone ids if we can identify the timezone.

    >>> import zoneinfo
    >>> from icalendar.timezone.tzid import tzids_from_tzinfo
    >>> tzids_from_tzinfo(zoneinfo.ZoneInfo("Arctic/Longyearbyen"))
    ('Arctic/Longyearbyen', 'Atlantic/Jan_Mayen', 'Europe/Berlin', 'Europe/Budapest', 'Europe/Copenhagen', 'Europe/Oslo', 'Europe/Stockholm', 'Europe/Vienna')
    >>> from dateutil.tz import gettz
    >>> tzids_from_tzinfo(gettz("Europe/Berlin"))
    ('Europe/Berlin', 'Arctic/Longyearbyen', 'Atlantic/Jan_Mayen', 'Europe/Budapest', 'Europe/Copenhagen', 'Europe/Oslo', 'Europe/Stockholm', 'Europe/Vienna')

    """  # The example might need to change if you recreate the lookup tree  # noqa: E501
    if tzinfo is None:
        return ()
    if isinstance(tzinfo, timezone):
        # fixed offset timezone with name
        return get_equivalent_tzids(tzinfo.tzname(None))
    if hasattr(tzinfo, "zone"):
        return get_equivalent_tzids(tzinfo.zone)  # pytz implementation
    if hasattr(tzinfo, "key"):
        return get_equivalent_tzids(tzinfo.key)  # ZoneInfo implementation
    if isinstance(tzinfo, tz._tzicalvtz):  # noqa: SLF001
        return get_equivalent_tzids(tzinfo._tzid)  # noqa: SLF001
    if isinstance(tzinfo, tz.tzstr):
        return get_equivalent_tzids(tzinfo._s)  # noqa: SLF001
    if hasattr(tzinfo, "_filename"):  # dateutil.tz.tzfile  # noqa: SIM102
        if DATEUTIL_ZONEINFO_PATH is not None:
            # tzfile('/usr/share/zoneinfo/Europe/Berlin')
            path = tzinfo._filename  # noqa: SLF001
            if path.startswith(str(DATEUTIL_ZONEINFO_PATH)):
                tzid = str(Path(path).relative_to(DATEUTIL_ZONEINFO_PATH))
                return get_equivalent_tzids(tzid)
            return get_equivalent_tzids(path)
    if isinstance(tzinfo, tz.tzutc):
        return get_equivalent_tzids("UTC")
    return ()


def tzid_from_tzinfo(tzinfo: Optional[tzinfo]) -> Optional[str]:
    """Retrieve the timezone id from the tzinfo object.

    Some timezones are equivalent.
    Thus, we might return one ID that is equivalent to others.
    """
    tzids = tzids_from_tzinfo(tzinfo)
    if "UTC" in tzids:
        return "UTC"
    if not tzids:
        return None
    return tzids[0]


def tzid_from_dt(dt: datetime) -> Optional[str]:
    """Retrieve the timezone id from the datetime object."""
    tzid = tzid_from_tzinfo(dt.tzinfo)
    if tzid is None:
        return dt.tzname()
    return tzid


_EQUIVALENT_IDS: dict[str, set[str]] = defaultdict(set)


def _add_equivalent_ids(value: tuple | dict | set):
    """This adds equivalent ids/

    As soon as one timezone implementation used claims their equivalence,
    they are considered equivalent.
    Have a look at icalendar.timezone.equivalent_timezone_ids.
    """
    if isinstance(value, set):
        for tzid in value:
            _EQUIVALENT_IDS[tzid].update(value)
    elif isinstance(value, tuple):
        _add_equivalent_ids(value[1])
    elif isinstance(value, dict):
        for value2 in value.values():
            _add_equivalent_ids(value2)
    else:
        raise TypeError(
            f"Expected tuple, dict or set, not {value.__class__.__name__}: {value!r}"
        )


_add_equivalent_ids(equivalent_timezone_ids_result.lookup)


def get_equivalent_tzids(tzid: str) -> tuple[str]:
    """This returns the tzids which are equivalent to this one."""
    ids = _EQUIVALENT_IDS.get(tzid, set())
    return (tzid,) + tuple(sorted(ids - {tzid}))


def is_utc(t: datetime | time | date | tzinfo) -> bool:
    """Whether this date is in UTC."""
    if is_date(t):
        return False
    tzid = tzid_from_dt(t) if hasattr(t, "tzinfo") else tzid_from_tzinfo(t)
    return tzid == "UTC"


__all__ = ["is_utc", "tzid_from_dt", "tzid_from_tzinfo", "tzids_from_tzinfo"]



"""Test conference properties."""

import pytest

from icalendar import Calendar, Event, Todo, vUri
from icalendar.prop.conference import Conference


def test_from_empty_uri():
    """Test creation from a URI."""
    conference = Conference.from_uri(vUri("https://chat.example.com/audio?id=123456"))
    assert conference.uri == "https://chat.example.com/audio?id=123456"
    assert conference.feature is None
    assert conference.label is None
    assert conference.language is None


def test_from_example_uri():
    """Create a conference with lots of values."""
    conference = Conference.from_uri(
        vUri(
            "https://chat.example.com/audio?id=123456",
            params={
                "FEATURE": "PHONE,MODERATOR",
                "LABEL": "Moderator dial-in",
                "LANGUAGE": "EN",
            },
        )
    )
    assert conference.uri == "https://chat.example.com/audio?id=123456"
    assert conference.feature == "PHONE,MODERATOR"
    assert conference.label == "Moderator dial-in"
    assert conference.language == "EN"


def test_to_uri():
    """Test creating a vURI."""
    uri = vUri(
        "https://chat.example.com/audio?id=123456",
        params={
            "FEATURE": "PHONE",
            "LABEL": "Moderator",
            "LANGUAGE": "DE",
            "VALUE": "URI",
        },
    )
    conference = Conference.from_uri(uri)
    new_uri = conference.to_uri()
    assert new_uri == uri
    assert new_uri.params == uri.params


def test_conference_list_params_serialization():
    """Ensure list parameters are correctly serialized to comma-separated strings."""
    conf = Conference(
        uri="https://example.com", feature=["AUDIO", "VIDEO"], label="Meeting room"
    )

    vuri = conf.to_uri()

    assert isinstance(vuri.params["FEATURE"], str)
    assert vuri.params["FEATURE"] == "AUDIO,VIDEO"
    assert vuri.params["LABEL"] == "Meeting room"


@pytest.fixture
def conference_1():
    """Fixture for a conference with a URI."""
    return Conference.from_uri(vUri("https://chat.example.com/audio?id=123456"))


@pytest.fixture
def conference_2():
    """Fixture for a conference with a URI."""
    return Conference(
        vUri("https://chat.example.com/audio?id=123456"),
        feature="PHONE",
        label="Moderator dial-in",
        language="EN",
    )


@pytest.fixture(params=[Event, Todo])
def component_class(request):
    """Fixture to create a Conference component."""
    return request.param


@pytest.fixture
def component(component_class):
    """Create a component."""
    return component_class()


def test_no_conferences(component):
    """No conferences by default."""
    assert component.conferences == []
    assert "CONFERENCE" not in component


def test_add_conference(component, conference_1):
    """Add a new conference."""
    component.conferences = [conference_1]
    assert component["CONFERENCE"] == conference_1.to_uri()


def test_add_multiple_conferences(component, conference_1, conference_2):
    """Add a new conference."""
    component.conferences = [conference_1, conference_2]
    assert component["CONFERENCE"] == [conference_1.to_uri(), conference_2.to_uri()]


def test_new_component_with_conferences(component_class, conference_1, conference_2):
    """Create a new component with conferences."""
    component = component_class.new(conferences=[conference_1, conference_2])
    assert component.conferences == [conference_1, conference_2]


def test_uri_in_ical_(conference_2, component):
    """The URI start be in the ical string."""
    component.conferences = [conference_2]
    ical_str = component.to_ical().decode("utf-8")
    assert "CONFERENCE;" in ical_str
    assert "/audio?id=123456" in ical_str
    assert "PHONE" in ical_str


def get_event(calendar: Calendar) -> Event:
    """Get the event directly"""
    return calendar.events[0]


def get_serialized_event(calendar: Calendar) -> Event:
    """Get the event directly"""
    cal = Calendar.from_ical(calendar.to_ical())
    return cal.events[0]


@pytest.fixture(params=[get_event, get_serialized_event])
def event_with_conferences(request, calendars):
    """Return the event from the file."""
    calendar = calendars.rfc_7986_conferences
    return request.param(calendar)


def test_conferences_from_file(event_with_conferences):
    """The conferences should be in the calendar."""
    assert len(event_with_conferences.conferences) == 5
    assert event_with_conferences.conferences[0].uri == "tel:+1-412-555-0123,,,654321"
    assert event_with_conferences.conferences[1].uri == "tel:+1-412-555-0123,,,555123"
    assert event_with_conferences.conferences[2].uri == "tel:+1-888-555-0456,,,555123"
    assert (
        event_with_conferences.conferences[3].uri
        == "xmpp:chat-123@conference.example.com"
    )
    assert (
        event_with_conferences.conferences[4].uri
        == "https://chat.example.com/audio?id=123456"
    )
    assert event_with_conferences.conferences[4].feature == ["AUDIO", "VIDEO"]
    assert event_with_conferences.conferences[4].label == "Attendee dial-in"
    assert event_with_conferences.conferences[4].language is None


def test_conference_from_string():
    """Create a conference from a string."""
    conference = Conference.from_uri("tel:+1-412-555-0123")
    assert conference.uri == "tel:+1-412-555-0123"
    assert conference.feature is None
    assert conference.label is None
    assert conference.language is None


def test_from_uri_string_adds_value_type():
    """Conference created from a string should add the value type."""
    conference = Conference.from_uri("http://asd")
    uri = conference.to_uri()
    assert uri.VALUE == "URI"



"""This tests the switch to different timezone implementations.

These are mostly located in icalendar.timezone.
"""

try:
    import pytz

    from icalendar.timezone.pytz import PYTZ
except ImportError:
    pytz = None
import copy
import pickle
from datetime import datetime

import pytest
from dateutil.rrule import MONTHLY, rrule
from dateutil.tz.tz import _tzicalvtz

from icalendar.timezone.zoneinfo import ZONEINFO, zoneinfo

if pytz:
    PYTZ_TIMEZONES = pytz.all_timezones
    TZP_ = [PYTZ(), ZONEINFO()]
    NEW_TZP_NAME = ["pytz", "zoneinfo"]
else:
    PYTZ_TIMEZONES = []
    TZP_ = [ZONEINFO()]
    NEW_TZP_NAME = ["zoneinfo"]


@pytest.mark.parametrize(
    "tz_name", PYTZ_TIMEZONES + list(zoneinfo.available_timezones())
)
@pytest.mark.parametrize("tzp_", TZP_)
def test_timezone_names_are_known(tz_name, tzp_):
    """Make sure that all timezones are understood."""
    if tz_name in ("Factory", "localtime"):
        pytest.skip()
    assert tzp_.knows_timezone_id(tz_name), (
        f"{tzp_.__class__.__name__} should know {tz_name}"
    )


@pytest.mark.parametrize("func", [pickle.dumps, copy.copy, copy.deepcopy])
@pytest.mark.parametrize(
    "obj",
    [
        _tzicalvtz("id"),
        rrule(freq=MONTHLY, count=4, dtstart=datetime(2028, 10, 1), cache=True),
    ],
)
def test_can_pickle_timezone(func, tzp, obj):
    """Check that we can serialize and copy timezones."""
    func(obj)


def test_copied_rrule_is_the_same():
    """When we copy an rrule, we want it to be the same after this."""
    r = rrule(freq=MONTHLY, count=4, dtstart=datetime(2028, 10, 1), cache=True)
    assert str(copy.deepcopy(r)) == str(r)


def test_tzp_properly_switches(tzp, tzp_name):
    """We want the default implementation to switch."""
    assert (tzp_name == "pytz") == tzp.uses_pytz()


def test_tzp_is_pytz_only(tzp, tzp_name, pytz_only):
    """We want the default implementation to switch."""
    assert tzp_name == "pytz"
    assert tzp.uses_pytz()


def test_cache_reuse_timezone_cache(tzp, timezones):
    """Make sure we do not cache the timezones twice and change them."""
    tzp.cache_timezone_component(timezones.pacific_fiji)
    tzp1 = tzp.timezone("custom_Pacific/Fiji")
    assert tzp1 is tzp.timezone("custom_Pacific/Fiji")
    tzp.cache_timezone_component(timezones.pacific_fiji)
    assert tzp1 is tzp.timezone("custom_Pacific/Fiji"), "Cache is not replaced."


@pytest.mark.parametrize("new_tzp_name", NEW_TZP_NAME)
def test_cache_is_emptied_when_tzp_is_switched(tzp, timezones, new_tzp_name):
    """Make sure we do not reuse the timezones created when we switch the provider."""
    tzp.cache_timezone_component(timezones.pacific_fiji)
    tz1 = tzp.timezone("custom_Pacific/Fiji")
    tzp.use(new_tzp_name)
    tzp.cache_timezone_component(timezones.pacific_fiji)
    tz2 = tzp.timezone("custom_Pacific/Fiji")
    assert tz1 is not tz2


def test_invalid_name(tzp):
    """Check that the provider name is OK."""
    provider = "invalid_provider"
    with pytest.raises(ValueError) as e:
        tzp.use(provider)
    # f"Unknown provider {provider}. Use 'pytz' or 'zoneinfo'."
    message = e.value.args[0]
    assert f"Unknown provider {provider}." in message
    assert "zoneinfo" in message
    assert "pytz" in message



# icalendar documentation build configuration file
import datetime
import os
import sys
from pathlib import Path

import icalendar

HERE = Path(__file__).parent
SRC = HERE.parent / "src"
sys.path.insert(0, str(SRC))  # update docs from icalendar source for livehtml

extensions = [
    "notfound.extension",
    "sphinx.ext.apidoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_reredirects",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",  # must be loaded after sphinx.ext.napoleon. See https://github.com/tox-dev/sphinx-autodoc-typehints/issues/15
]
source_suffix = {".rst": "restructuredtext"}
master_doc = "index"

project = "icalendar"
this_year = datetime.date.today().year  # noqa: DTZ011
copyright = f"{this_year}, Plone Foundation"  # noqa: A001

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# Define the json_url for our version switcher.
json_url = "https://icalendar.readthedocs.io/en/latest/_static/version-switcher.json"

# Define the version we use for matching in the version switcher.
version_match = os.environ.get("READTHEDOCS_VERSION")
release = icalendar.__version__

# If READTHEDOCS_VERSION doesn't exist, we're not on RTD
# If it's an integer, we're in a PR build and the version isn't correct.
# If it's "latest", change to "dev" in the version switcher.
if not version_match or version_match.isdigit() or version_match == "latest":
    # For local development, infer the version to match from the package.
    if "a" in release or "b" in release or "rc" in release or "dev" in release:
        # Override the fully qualified URL in dev mode.
        json_url = "_static/version-switcher.json"
        version_match = "dev"
    else:
        version_match = f"v{release}"
elif version_match == "stable":
    version_match = f"v{release}"

exclude_patterns = [
    "reference/api/modules.rst",
]
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/collective/icalendar",
            "icon": "fa-brands fa-square-github",
            "type": "fontawesome",
            "attributes": {
                "target": "_blank",
                "rel": "noopener me",
                "class": "nav-link custom-fancy-css",
            },
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/icalendar",
            "icon": "fa-custom fa-pypi",
            "type": "fontawesome",
            "attributes": {
                "target": "_blank",
                "rel": "noopener me",
                "class": "nav-link custom-fancy-css",
            },
        },
    ],
    "logo": {"text": "icalendar"},
    "navbar_start": ["navbar-logo", "version-switcher"],
    "navigation_with_keys": True,
    "search_bar_text": "Search",
    "show_nav_level": 2,
    "show_toc_level": 2,
    "show_version_warning_banner": True,
    "switcher": {
        "json_url": json_url,
        "version_match": version_match,
    },
    "use_edit_page_button": True,
}
html_context = {
    #     "github_url": "https://github.com", # or your GitHub Enterprise site
    "github_user": "collective",
    "github_repo": "icalendar",
    "github_version": "main",
    "doc_path": "docs",
}
html_static_path = [
    "_static",
]
html_js_files = [
    ("js/custom-icons.js", {"defer": "defer"}),
]
napoleon_use_param = True
pygments_style = "sphinx"

# -- Intersphinx configuration ----------------------------------

# This extension can generate automatic links to the documentation of objects
# in other projects. Usage is simple: whenever Sphinx encounters a
# cross-reference that has no matching target in the current documentation set,
# it looks for targets in the documentation sets configured in
# intersphinx_mapping. A reference like :py:class:`zipfile.ZipFile` can then
# linkto the Python documentation for the ZipFile class, without you having to
# specify where it is located exactly.
#
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}


# -- linkcheck configuration ----------------------------------
# Ignore localhost
linkcheck_ignore = [
    # Ignore pages that require authentication
    r"https://app.readthedocs.org/dashboard/icalendar/users/create/",
    r"https://github.com/collective/icalendar/fork",
    r"https://github.com/collective/icalendar/settings/",
    r"https://groups.google.com/g/icalendar-coc/",
    r"https://pypi.org/manage/project/icalendar/collaboration/",
    # Ignore specific anchors
    r"https://github.com/collective/icalendar/blob/main/CODE_OF_CONDUCT.md#enforcement",
    r"https://github.com/collective/icalendar/blob/main/README.rst#related-projects",
    r"https://up-for-grabs.net/#/filters",
    # Ignore links that are unstable
    r"https://www.unicode.org/cldr/cldr-aux/charts/29/supplemental/zone_tzid.html",
]
linkcheck_anchors = True
linkcheck_timeout = 5
linkcheck_retries = 1

# -- Options for sphinx-notfound-page ----------------------------------
# https://sphinx-notfound-page.readthedocs.io/en/latest/configuration.html
notfound_template = "404.html"


# -- sphinx-reredirects configuration ----------------------------------
# https://documatt.com/sphinx-reredirects/usage.html
redirects = {
    "about": "index.html",
    "api": "reference/api/icalendar.html",
    "changelog": "reference/changelog.html",
    "cli": "how-to/cli.html",
    "credits": "contribute/credits.html",
    "contributing": "contribute/index.html",
    "install": "how-to/install.html",
    "license": "https://github.com/collective/icalendar/blob/main/LICENSE.rst",
    "maintenance": "contribute/maintenance.html",
    "security": "https://github.com/collective/icalendar/blob/main/SECURITY.md",
    "usage": "how-to/usage.html",
    }

man_pages = [("index", "icalendar", "icalendar Documentation", ["Plone Foundation"], 1)]

htmlhelp_basename = "icalendardoc"



"""This implemensts RFC 6868.

There are only some changes to parameters needed.
"""

import os
from typing import TYPE_CHECKING

import pytest

from icalendar.cal.calendar import Calendar
from icalendar.parser import dquote, rfc_6868_escape, rfc_6868_unescape

if TYPE_CHECKING:
    from icalendar import vCalAddress, vText


param_duplicate = pytest.mark.parametrize(
    "duplicate", [lambda x: x, lambda cal: Calendar.from_ical(cal.to_ical())]
)


@param_duplicate
def test_rfc_6868_example(calendars, duplicate):
    """Check the example from the RFC."""
    cal: Calendar = duplicate(calendars.rfc_6868)
    attendee: vCalAddress = cal.events[0]["attendee"]
    assert attendee.name == 'George Herman "Babe" Ruth'


@param_duplicate
def test_all_parameters(calendars, duplicate):
    """Check that all examples get decoded correctly."""
    cal: Calendar = duplicate(calendars.rfc_6868)
    param = cal["X-PARAM"].params["ALL"]
    assert param == '^"\n'


@param_duplicate
def test_unknown_character(calendars, duplicate):
    """if a ^ (U+005E) character is followed by any character other than
    the ones above, parsers MUST leave both the ^ and the following
    character in place"""
    cal: Calendar = duplicate(calendars.rfc_6868)
    param = cal["X-PARAM"].params["UNKNOWN"]
    assert param == "^a^ ^asd"


@pytest.mark.parametrize(
    ("text", "expected"),
    [
        ("^a", "^a"),
        ("^^", "^"),
        # ("^n", ),  # see other test
        ("^'", '"'),
        ("^^a^b", "^a^b"),
    ],
)
@pytest.mark.parametrize(
    "modify", [lambda x: x, lambda x: x * 10, lambda x: f"asd{x}aaA^AA"]
)
def test_unescape(text, expected, modify):
    """Check unescaping."""
    result = rfc_6868_unescape(modify(text))
    assert result == modify(expected)


@pytest.mark.parametrize("newline", ["\n", "\r\n", "\r", os.linesep])
def test_unescape_newline(newline, monkeypatch):
    """Unescape the newline."""
    monkeypatch.setattr(os, "linesep", newline)
    assert rfc_6868_unescape("^n") == newline


param_values = pytest.mark.parametrize(
    ("text", "expected"),
    [
        ("", ""),
        ("^", "^^"),
        ("^n", "^^n"),
        ("This text\n has\r several\r\n lines", "This text^n has^n several^n lines"),
        ('Call me "Harry".', "Call me ^'Harry^'."),
    ],
)


@param_values
def test_escape_rfc_6868(text, expected):
    """Check that we can escape the content with special characters."""
    escaped = rfc_6868_escape(text)
    assert escaped == expected, f"{escaped!r} == {expected!r}"
    assert rfc_6868_escape(rfc_6868_unescape(escaped)) == expected


@param_values
def test_escaping_parameters(text, expected):
    cal = Calendar()
    cal.add("X-Param", "asd")
    param: vText = cal["X-PARAM"]
    param.params["RFC6868"] = text
    ical = cal.to_ical().decode()
    print(ical)
    assert "X-PARAM;RFC6868=" + dquote(expected) in ical


def test_encode_example_again(calendars):
    """The example file should yield its content again."""
    cal: Calendar = calendars.rfc_6868
    again = Calendar.from_ical(cal.to_ical())
    assert cal == again
    assert cal.to_ical() == again.to_ical()



import datetime

import pytest


@pytest.mark.parametrize(
    ("field", "expected_value"),
    [
        ("PRODID", "-//Plönë.org//NONSGML plone.app.event//EN"),
        ("X-WR-CALDESC", "test non ascii: äöü ÄÖÜ €"),
    ],
)
def test_calendar_from_ical_respects_unicode(field, expected_value, calendars):
    cal = calendars.calendar_with_unicode
    assert cal[field].to_ical().decode("utf-8") == expected_value


@pytest.mark.parametrize(
    ("test_input", "field", "expected_value"),
    [
        ("event_with_unicode_fields", "SUMMARY", "Non-ASCII Test: ÄÖÜ äöü €"),
        (
            "event_with_unicode_fields",
            "DESCRIPTION",
            "icalendar should be able to handle non-ascii: €äüöÄÜÖ.",
        ),
        ("event_with_unicode_fields", "LOCATION", "Tribstrül"),
        # Non-unicode characters in summary
        # https://github.com/collective/icalendar/issues/64
        ("issue_64_event_with_non_ascii_summary", "SUMMARY", "åäö"),
        # Unicode characters in summary
        ("issue_64_event_with_ascii_summary", "SUMMARY", "abcdef"),
    ],
)
def test_event_from_ical_respects_unicode(test_input, field, expected_value, events):
    event = events[test_input]
    assert event[field].to_ical().decode("utf-8") == expected_value


@pytest.mark.parametrize(
    ("test_input", "expected_output"),
    [
        # chokes on umlauts in ORGANIZER
        # https://github.com/collective/icalendar/issues/101
        ("issue_101_icalendar_chokes_on_umlauts_in_organizer", "acme, ädmin"),
        ("event_with_unicode_organizer", "Джон Доу"),
    ],
)
def test_events_parameter_unicoded(events, test_input, expected_output):
    assert events[test_input]["ORGANIZER"].params["CN"] == expected_output


def test_parses_event_with_non_ascii_tzid_issue_237(calendars, in_timezone):
    """Issue #237 - Fail to parse timezone with non-ascii TZID
    see https://github.com/collective/icalendar/issues/237
    """
    start = calendars.issue_237_fail_to_parse_timezone_with_non_ascii_tzid.walk(
        "VEVENT"
    )[0].decoded("DTSTART")
    expected = in_timezone(datetime.datetime(2017, 5, 11, 13, 30), "America/Sao_Paulo")
    assert not calendars.issue_237_fail_to_parse_timezone_with_non_ascii_tzid.errors
    assert start == expected


def test_parses_timezone_with_non_ascii_tzid_issue_237(timezones):
    """Issue #237 - Fail to parse timezone with non-ascii TZID
    see https://github.com/collective/icalendar/issues/237
    """
    assert timezones.issue_237_brazilia_standard["tzid"] == "(UTC-03:00) Brasília"


@pytest.mark.parametrize("timezone_name", ["standard", "daylight"])
def test_parses_timezone_with_non_ascii_tzname_issue_273(timezones, timezone_name):
    """Issue #237 - Fail to parse timezone with non-ascii TZID
    see https://github.com/collective/icalendar/issues/237
    """
    assert (
        timezones.issue_237_brazilia_standard.walk(timezone_name)[0]["TZNAME"]
        == f"Brasília {timezone_name}"
    )


def test_broken_property(calendars):
    """
    Test if error messages are encoded properly.
    """
    for event in calendars.broken_ical.walk("vevent"):
        assert len(event.errors) == 1, "Not the right amount of errors."
        error = event.errors[0][1]
        assert error.startswith("Content line could not be parsed into parts")


def test_apple_xlocation(calendars):
    """
    Test if we support base64 encoded binary data in parameter values.
    """
    for event in calendars.x_location.walk("vevent"):
        assert len(event.errors) == 0, "Got too many errors"



"""Test the equality and inequality of components."""

import contextlib
import copy

try:
    from pytz import UnknownTimeZoneError
except ImportError:

    class UnknownTimeZoneError(Exception):
        pass


from datetime import date, datetime, time, timedelta

import pytest

from icalendar import (
    Component,
    vBinary,
    vBoolean,
    vCategory,
    vDate,
    vDatetime,
    vDDDLists,
    vDDDTypes,
    vDuration,
    vGeo,
    vPeriod,
    vRecur,
    vText,
    vTime,
)


def assert_equal(actual_value, expected_value):
    """Make sure both values are equal"""
    assert actual_value == expected_value
    assert expected_value == actual_value


def assert_not_equal(actual_value, expected_value):
    """Make sure both values are not equal"""
    assert actual_value != expected_value
    assert expected_value != actual_value


def test_parsed_calendars_are_equal_if_parsed_again(source_file, tzp):
    """Ensure that a calendar equals the same calendar.

    source -> calendar -> ics -> same calendar
    """
    if source_file.source_file in ("rfc_7265_example_2.jcal",):
        pytest.skip("TODO: Fix in https://github.com/collective/icalendar/issues/1008.")
    copy_of_calendar = source_file.__class__.from_ical(source_file.to_ical())
    assert_equal(copy_of_calendar, source_file)


def test_parsed_calendars_are_equal_if_parsed_again_jcal(source_file, tzp):
    """Ensure that a calendar equals the same calendar.

    source -> calendar -> jcal -> same calendar
    """
    copy_of_calendar = Component.from_jcal(source_file.to_jcal())
    assert_equal(copy_of_calendar, source_file)


def test_parsed_calendars_are_equal_if_from_same_source(ics_file, tzp):
    """Ensure that a calendar equals the same calendar.

    ics -> calendar
    ics -> same calendar
    """
    cal1 = ics_file.__class__.from_ical(ics_file.raw_ics)
    cal2 = ics_file.__class__.from_ical(ics_file.raw_ics)
    assert_equal(cal1, cal2)


def test_parsed_calendars_are_equal_if_from_same_source_jcal(jcal_file, tzp):
    """Ensure that a calendar equals the same calendar.

    jcal -> calendar
    jcal -> same calendar
    """
    cal1 = Component.from_jcal(jcal_file.raw_jcal)
    cal2 = Component.from_jcal(jcal_file.raw_jcal)
    assert_equal(cal1, cal2)


def test_copies_are_equal(source_file, tzp):
    """Ensure that copies are equal."""
    copy1 = source_file.copy()
    copy1.subcomponents = source_file.subcomponents
    copy2 = source_file.copy()
    copy2.subcomponents = source_file.subcomponents[:]
    assert_equal(copy1, copy2)
    assert_equal(copy1, source_file)
    assert_equal(copy2, source_file)


def test_copy_does_not_copy_subcomponents(calendars, tzp):
    """If we copy the subcomponents, assumptions around copies will be broken."""
    assert calendars.timezoned.subcomponents
    assert not calendars.timezoned.copy().subcomponents


def test_deep_copies_are_equal(source_file, tzp):
    """Ensure that deep copies are equal.

    Ignore errors when a custom time zone is used.
    This is still covered by the parsing test.
    """
    if (
        source_file.source_file == "issue_722_timezone_transition_ambiguity.ics"
        and tzp.uses_zoneinfo()
    ):
        pytest.skip("This test fails for now.")
    with contextlib.suppress(UnknownTimeZoneError):
        assert_equal(copy.deepcopy(source_file), copy.deepcopy(source_file))
    with contextlib.suppress(UnknownTimeZoneError):
        assert_equal(copy.deepcopy(source_file), source_file)


def test_vGeo():
    """Check the equality of vGeo."""
    assert_equal(vGeo(("100", "12.33")), vGeo(("100.00", "12.330")))
    assert_not_equal(vGeo(("100", "12.331")), vGeo(("100.00", "12.330")))
    assert_not_equal(vGeo(("10", "12.33")), vGeo(("100.00", "12.330")))


def test_vBinary():
    assert_equal(vBinary("asd"), vBinary("asd"))
    assert_not_equal(vBinary("asdf"), vBinary("asd"))


def test_vBoolean():
    assert_equal(vBoolean.from_ical("TRUE"), vBoolean.from_ical("TRUE"))
    assert_equal(vBoolean.from_ical("FALSE"), vBoolean.from_ical("FALSE"))
    assert_not_equal(vBoolean.from_ical("TRUE"), vBoolean.from_ical("FALSE"))


def test_vCategory():
    assert_equal(vCategory("HELLO"), vCategory("HELLO"))
    assert_equal(vCategory(["a", "b"]), vCategory(["a", "b"]))
    assert_not_equal(vCategory(["a", "b"]), vCategory(["a", "b", "c"]))


def test_vText():
    assert_equal(vText("HELLO"), vText("HELLO"))
    assert_not_equal(vText("HELLO1"), vText("HELLO"))


@pytest.mark.parametrize(
    ("v_type", "v1", "v2"),
    [
        (vDatetime, datetime(2023, 11, 1, 10, 11), datetime(2023, 11, 1, 10, 10)),
        (vDate, date(2023, 11, 1), date(2023, 10, 31)),
        (vDuration, timedelta(3, 11, 1), timedelta(23, 10, 31)),
        (
            vPeriod,
            (datetime(2023, 11, 1, 10, 11), timedelta(3, 11, 1)),
            (datetime(2023, 11, 1, 10, 11), timedelta(23, 10, 31)),
        ),
        (
            vPeriod,
            (datetime(2023, 11, 1, 10, 1), timedelta(3, 11, 1)),
            (datetime(2023, 11, 1, 10, 11), timedelta(3, 11, 1)),
        ),
        (
            vPeriod,
            (datetime(2023, 11, 1, 10, 1), datetime(2023, 11, 1, 10, 3)),
            (datetime(2023, 11, 1, 10, 1), datetime(2023, 11, 1, 10, 2)),
        ),
        (vTime, time(10, 10, 10), time(10, 10, 11)),
    ],
)
@pytest.mark.parametrize("eq", ["==", "!="])
@pytest.mark.parametrize("cls1", [0, 1])
@pytest.mark.parametrize("cls2", [0, 1])
@pytest.mark.parametrize("hash", [lambda x: x, hash])
def test_vDDDTypes_and_others(v_type, v1, v2, cls1, cls2, eq, hash):  # noqa: A002
    """Check equality and inequality."""
    t1 = (v_type, vDDDTypes)[cls1]
    t2 = (v_type, vDDDTypes)[cls2]
    if eq == "==":
        assert hash(v1) == hash(v1)
        assert hash(t1(v1)) == hash(t2(v1))
        assert hash(t1(v1)) == hash(t2(v1))
    else:
        assert hash(v1) != hash(v2)
        assert hash(t1(v1)) != hash(t2(v2))


def test_repr_vDDDTypes():
    assert "vDDDTypes" in repr(vDDDTypes(timedelta(3, 11, 1)))


vDDDLists_examples = [  # noqa: N816
    vDDDLists([]),
    vDDDLists([datetime(2023, 11, 1, 10, 1)]),
    vDDDLists([datetime(2023, 11, 1, 10, 1), date(2023, 11, 1)]),
]


@pytest.mark.parametrize("l1", vDDDLists_examples)
@pytest.mark.parametrize("l2", vDDDLists_examples)
def test_vDDDLists(l1, l2):
    """Check the equality functions of vDDDLists."""
    equal = l1 is l2
    l2 = copy.deepcopy(l2)
    assert equal == (l1 == l2)
    assert equal != (l1 != l2)


@pytest.mark.parametrize("index", [1, 2])
def test_rfc_7265_equivalence_of_example_from_appendix(calendars, index):
    """jcal and ical should be the same"""
    ical = calendars[f"rfc_7265_appendix_example_{index}_ical"]
    jcal = calendars[f"rfc_7265_appendix_example_{index}_jcal"]
    assert_equal(ical.to_jcal(), jcal.to_jcal())
    assert_equal(ical, jcal)


@pytest.mark.parametrize(
    ("v1", "v2", "equal"),
    [
        ({}, {}, True),
        ({"BYDAY": "1SU"}, {}, False),
        ({"BYDAY": "1SU"}, {"BYDAY": "1SU"}, True),
        ({"BYDAY": "1SU"}, {"BYDAY": ["1SU"]}, True),
        ({"byday": "1SU"}, {"BYDAY": "1SU"}, True),
        ({"byday": "1SU", "count": 1}, {"BYDAY": "1SU"}, False),
        ({"count": 1}, {"BYDAY": "1SU"}, False),
    ],
)
def test_v_recur_equal(v1, v2, equal):
    """Check the equality functions of vRecur."""
    recur1 = vRecur(v1)
    recur2 = vRecur(v2)
    if equal:
        assert_equal(recur1, recur2)
    else:
        assert_not_equal(recur1, recur2)



"""This implementes the VAVAILABILITY component.

This is specified in :rfc:`7953`.
"""

from __future__ import annotations

import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Optional, Sequence

from icalendar.attr import (
    CONCEPTS_TYPE_SETTER,
    LINKS_TYPE_SETTER,
    RELATED_TO_TYPE_SETTER,
    busy_type_property,
    categories_property,
    class_property,
    contacts_property,
    description_property,
    duration_property,
    location_property,
    organizer_property,
    priority_property,
    rfc_7953_dtend_property,
    rfc_7953_dtstart_property,
    rfc_7953_duration_property,
    rfc_7953_end_property,
    sequence_property,
    summary_property,
    url_property,
)
from icalendar.cal.examples import get_example
from icalendar.error import InvalidCalendar

from .component import Component

if TYPE_CHECKING:
    from datetime import date

    from icalendar.cal.available import Available
    from icalendar.enums import BUSYTYPE, CLASS
    from icalendar.prop import vCalAddress


class Availability(Component):
    """VAVAILABILITY component from :rfc:`7953`.

    This provides a grouping of component properties and
    subcomponents that describe the availability associated with a
    calendar user.

    Description:
        A "VAVAILABILITY" component indicates a period of time
        within which availability information is provided.  A
        "VAVAILABILITY" component can specify a start time and an end time
        or duration.  If "DTSTART" is not present, then the start time is
        unbounded.  If "DTEND" or "DURATION" are not present, then the end
        time is unbounded.  Within the specified time period, availability
        defaults to a free-busy type of "BUSY-UNAVAILABLE" (see
        Section 3.2), except for any time periods corresponding to
        "AVAILABLE" subcomponents.

        "AVAILABLE" subcomponents are used to indicate periods of free
        time within the time range of the enclosing "VAVAILABILITY"
        component.  "AVAILABLE" subcomponents MAY include recurrence
        properties to specify recurring periods of time, which can be
        overridden using normal iCalendar recurrence behavior (i.e., use
        of the "RECURRENCE-ID" property).

        If specified, the "DTSTART" and "DTEND" properties in
        "VAVAILABILITY" components and "AVAILABLE" subcomponents MUST be
        "DATE-TIME" values specified as either the date with UTC time or
        the date with local time and a time zone reference.

        The iCalendar object containing the "VAVAILABILITY" component MUST
        contain appropriate "VTIMEZONE" components corresponding to each
        unique "TZID" parameter value used in any DATE-TIME properties in
        all components, unless [RFC7809] is in effect.

        When used to publish available time, the "ORGANIZER" property
        specifies the calendar user associated with the published
        available time.

        If the "PRIORITY" property is specified in "VAVAILABILITY"
        components, it is used to determine how that component is combined
        with other "VAVAILABILITY" components.  See Section 4.

        Other calendar properties MAY be specified in "VAVAILABILITY" or
        "AVAILABLE" components and are considered attributes of the marked
        block of time.  Their usage is application specific.  For example,
        the "LOCATION" property might be used to indicate that a person is
        available in one location for part of the week and a different
        location for another part of the week (but see Section 9 for when
        it is appropriate to add additional data like this).

    Example:
        The following is an example of a "VAVAILABILITY" calendar
        component used to represent the availability of a user, always
        available Monday through Friday, 9:00 am to 5:00 pm in the
        America/Montreal time zone:

        .. code-block:: text

            BEGIN:VAVAILABILITY
            ORGANIZER:mailto:bernard@example.com
            UID:0428C7D2-688E-4D2E-AC52-CD112E2469DF
            DTSTAMP:20111005T133225Z
            BEGIN:AVAILABLE
            UID:34EDA59B-6BB1-4E94-A66C-64999089C0AF
            SUMMARY:Monday to Friday from 9:00 to 17:00
            DTSTART;TZID=America/Montreal:20111002T090000
            DTEND;TZID=America/Montreal:20111002T170000
            RRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR
            END:AVAILABLE
            END:VAVAILABILITY

        You can get the same example from :meth:`example`:

        .. code-block: pycon

            >>> from icalendar import Availability
            >>> a = Availability.example()
            >>> a.organizer
            vCalAddress('mailto:bernard@example.com')

        The following is an example of a "VAVAILABILITY" calendar
        component used to represent the availability of a user available
        Monday through Thursday, 9:00 am to 5:00 pm, at the main office,
        and Friday, 9:00 am to 12:00 pm, in the branch office in the
        America/Montreal time zone between October 2nd and December 2nd
        2011:

        .. code-block:: text

            BEGIN:VAVAILABILITY
            ORGANIZER:mailto:bernard@example.com
            UID:84D0F948-7FC6-4C1D-BBF3-BA9827B424B5
            DTSTAMP:20111005T133225Z
            DTSTART;TZID=America/Montreal:20111002T000000
            DTEND;TZID=America/Montreal:20111202T000000
            BEGIN:AVAILABLE
            UID:7B33093A-7F98-4EED-B381-A5652530F04D
            SUMMARY:Monday to Thursday from 9:00 to 17:00
            DTSTART;TZID=America/Montreal:20111002T090000
            DTEND;TZID=America/Montreal:20111002T170000
            RRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH
            LOCATION:Main Office
            END:AVAILABLE
            BEGIN:AVAILABLE
            UID:DF39DC9E-D8C3-492F-9101-0434E8FC1896
            SUMMARY:Friday from 9:00 to 12:00
            DTSTART;TZID=America/Montreal:20111006T090000
            DTEND;TZID=America/Montreal:20111006T120000
            RRULE:FREQ=WEEKLY
            LOCATION:Branch Office
            END:AVAILABLE
            END:VAVAILABILITY

        For more examples, have a look at :rfc:`5545`.

    """

    name = "VAVAILABILITY"

    canonical_order = (
        "DTSTART",
        "DTEND",
        "DURATION",
        "DTSTAMP",
        "UID",
        "SEQUENCE",
        "SUMMARY",
        "DESCRIPTION",
        "ORGANIZER",
    )

    required = (
        "DTSTART",
        "DTSTAMP",
        "UID",
    )

    singletons = (
        "DTSTAMP",
        "UID",
        "BUSYTYPE",
        "CLASS",
        "CREATED",
        "DESCRIPTION",
        "DTSTART",
        "LAST-MODIFIED",
        "LOCATION",
        "ORGANIZER",
        "PRIORITY",
        "SEQUENCE",
        "SUMMARY",
        "URL",
        "DTEND",
        "DURATION",
    )

    exclusive = (
        "DTEND",
        "DURATION",
    )

    organizer = organizer_property
    busy_type = busy_type_property
    summary = summary_property
    description = description_property
    sequence = sequence_property
    classification = class_property
    url = url_property
    location = location_property
    categories = categories_property
    priority = priority_property
    contacts = contacts_property

    start = DTSTART = rfc_7953_dtstart_property
    DTEND = rfc_7953_dtend_property
    DURATION = duration_property("Availability")
    duration = rfc_7953_duration_property
    end = rfc_7953_end_property

    @property
    def available(self) -> list[Available]:
        """All VAVAILABLE sub-components.

        This is a shortcut to get all VAVAILABLE sub-components.
        Modifications do not change the calendar.
        Use :py:meth:`Component.add_component`.
        """
        return self.walk("AVAILABLE")

    @classmethod
    def new(
        cls,
        /,
        busy_type: Optional[BUSYTYPE] = None,
        categories: Sequence[str] = (),
        comments: list[str] | str | None = None,
        components: Sequence[Available] | None = (),
        concepts: CONCEPTS_TYPE_SETTER = None,
        contacts: list[str] | str | None = None,
        created: Optional[date] = None,
        classification: Optional[CLASS] = None,
        description: Optional[str] = None,
        end: Optional[datetime] = None,
        last_modified: Optional[date] = None,
        links: LINKS_TYPE_SETTER = None,
        location: Optional[str] = None,
        organizer: Optional[vCalAddress | str] = None,
        priority: Optional[int] = None,
        refids: list[str] | str | None = None,
        related_to: RELATED_TO_TYPE_SETTER = None,
        sequence: Optional[int] = None,
        stamp: Optional[date] = None,
        start: Optional[datetime] = None,
        summary: Optional[str] = None,
        uid: Optional[str | uuid.UUID] = None,
        url: Optional[str] = None,
    ):
        """Create a new event with all required properties.

        This creates a new Availability in accordance with :rfc:`7953`.

        Arguments:
            busy_type: The :attr:`busy_type` of the availability.
            categories: The :attr:`categories` of the availability.
            classification: The :attr:`classification` of the availability.
            comments: The :attr:`~icalendar.Component.comments` of the availability.
            concepts: The :attr:`~icalendar.Component.concepts` of the availability.
            contacts: The :attr:`contacts` of the availability.
            created: The :attr:`~icalendar.Component.created` of the availability.
            description: The :attr:`description` of the availability.
            end: The :attr:`end` of the availability.
            last_modified: The :attr:`~icalendar.Component.last_modified` of the
                availability.
            links: The :attr:`~icalendar.Component.links` of the availability.
            location: The :attr:`location` of the availability.
            organizer: The :attr:`organizer` of the availability.
            refids: :attr:`~icalendar.Component.refids` of the availability.
            related_to: :attr:`~icalendar.Component.related_to` of the availability.
            sequence: The :attr:`sequence` of the availability.
            stamp: The :attr:`~icalendar.Component.stamp` of the availability.
                If None, this is set to the current time.
            start: The :attr:`start` of the availability.
            summary: The :attr:`summary` of the availability.
            uid: The :attr:`uid` of the availability.
                If None, this is set to a new :func:`uuid.uuid4`.
            url: The :attr:`url` of the availability.

        Returns:
            :class:`Availability`

        Raises:
            InvalidCalendar: If the content is not valid according to :rfc:`7953`.

        .. warning:: As time progresses, we will be stricter with the validation.
        """
        availability: Availability = super().new(
            stamp=stamp if stamp is not None else cls._utc_now(),
            created=created,
            comments=comments,
            last_modified=last_modified,
            links=links,
            related_to=related_to,
            refids=refids,
            concepts=concepts,
        )
        availability.summary = summary
        availability.description = description
        availability.uid = uid if uid is not None else uuid.uuid4()
        availability.sequence = sequence
        availability.categories = categories
        availability.classification = classification
        availability.url = url
        availability.busy_type = busy_type
        availability.organizer = organizer
        availability.location = location
        availability.priority = priority
        availability.contacts = contacts
        for subcomponent in components:
            availability.add_component(subcomponent)
        if cls._validate_new:
            if start is not None and (
                not isinstance(start, datetime) or start.tzinfo is None
            ):
                raise InvalidCalendar(
                    "Availability start must be a datetime with a timezone"
                )
            if end is not None and (
                not isinstance(end, datetime) or end.tzinfo is None
            ):
                raise InvalidCalendar(
                    "Availability end must be a datetime with a timezone"
                )
            availability._validate_start_and_end(start, end)
        availability.start = start
        availability.end = end
        return availability

    @classmethod
    def example(cls, name: str = "rfc_7953_1") -> Availability:
        """Return the calendar example with the given name."""
        return cls.from_ical(get_example("availabilities", name))


__all__ = ["Availability"]



"""The GAP parameter can be defined on RELATED-TO properties."""

from datetime import timedelta
from typing import TYPE_CHECKING

import pytest

from icalendar import Journal
from icalendar.error import InvalidCalendar
from icalendar.prop import vText, vUid, vUri

if TYPE_CHECKING:
    from icalendar import Calendar


@pytest.fixture
def component():
    """Return a basic component for testing."""
    return Journal()


def test_get_gap_example(calendars):
    """Get the gap from an example calendar."""
    calendar: Calendar = calendars.rfc_9253_gap
    event1 = calendar.events[0]
    event2 = calendar.events[1]
    assert event1.related_to == []
    relation: vUri = event2.related_to[0]
    assert relation.params["GAP"]
    assert relation.GAP == timedelta(weeks=1)
    assert relation.uid == "1"


@pytest.fixture(params=[vText, vUid, vUri])
def related_to_prop(request):
    """The RELATED-TO property classes to test."""
    return request.param("uid-text-or-url")


@pytest.mark.parametrize(
    ("gap_input", "value"),
    [
        (timedelta(days=3), timedelta(days=3)),
        ("P15D", timedelta(days=15)),
        ("-PT6H", timedelta(hours=-6)),
    ],
)
def test_get_and_set_gap(gap_input, value, related_to_prop):
    """Check that we can get the GAP parameter."""
    related_to_prop.GAP = gap_input
    assert related_to_prop.GAP == value
    assert related_to_prop.params["GAP"]


def test_serialize(component, related_to_prop):
    """Check that we can serialize the GAP parameter."""
    related_to_prop.GAP = timedelta(days=10)
    component.add("RELATED-TO", related_to_prop)
    ical = component.to_ical().decode()
    assert "GAP=P10D" in ical


def test_delete_gap(related_to_prop):
    """Check that we can delete the GAP parameter."""
    related_to_prop.GAP = timedelta(days=5)
    del related_to_prop.GAP
    assert related_to_prop.GAP is None


def test_delete_gap_by_setting_to_None(related_to_prop):
    """Check that we can delete the GAP parameter."""
    related_to_prop.GAP = timedelta(days=5)
    related_to_prop.GAP = None
    assert related_to_prop.GAP is None


def test_set_invalid_value(related_to_prop):
    """Check setting invalid values."""
    with pytest.raises(ValueError):
        related_to_prop.GAP = "invalid"
    with pytest.raises(TypeError):
        related_to_prop.GAP = object()


def test_get_malformed_value(related_to_prop):
    """What if the value is a string?"""
    related_to_prop.params["GAP"] = "invalid"
    with pytest.raises(InvalidCalendar):
        related_to_prop.GAP  # noqa: B018


def test_gap_invalid_value(related_to_prop):
    """Other value types could be there but are problematic."""
    related_to_prop.params["GAP"] = object()
    with pytest.raises(TypeError):
        related_to_prop.GAP  # noqa: B018



"""Compute the times and states of alarms.

This takes different calendar software into account and the RFC 9074 (Alarm Extension).

- RFC 9074 defines an ACKNOWLEDGED property in the VALARM.
- Outlook does not export VALARM information.
- Google Calendar uses the DTSTAMP to acknowledge the alarms.
- Thunderbird snoozes the alarms with a X-MOZ-SNOOZE-TIME attribute in the event.
- Thunderbird acknowledges the alarms with a X-MOZ-LASTACK attribute in the event.
- Etar deletes alarms that are acknowledged.
- Nextcloud's Webinterface does not do anything with the alarms when the time passes.
"""

from __future__ import annotations

from datetime import date, timedelta, tzinfo
from typing import TYPE_CHECKING, Generator, Optional, Union

from icalendar.cal.event import Event
from icalendar.cal.todo import Todo
from icalendar.error import (
    ComponentEndMissing,
    ComponentStartMissing,
    IncompleteAlarmInformation,
    LocalTimezoneMissing,
)
from icalendar.timezone import tzp
from icalendar.tools import is_date, normalize_pytz, to_datetime

if TYPE_CHECKING:
    from datetime import datetime

    from icalendar.cal.alarm import Alarm

Parent = Union[Event, Todo]


class AlarmTime:
    """An alarm time with all the information."""

    def __init__(
        self,
        alarm: Alarm,
        trigger: datetime,
        acknowledged_until: Optional[datetime] = None,
        snoozed_until: Optional[datetime] = None,
        parent: Optional[Parent] = None,
    ):
        """Create a new AlarmTime.

        alarm
            the Alarm component

        trigger
            a date or datetime at which to trigger the alarm

        acknowledged_until
            an optional datetime in UTC until when all alarms
            have been acknowledged

        snoozed_until
            an optional datetime in UTC until which all alarms of
            the same parent are snoozed

        parent
            the optional parent component the alarm refers to

        local_tzinfo
            the local timezone that events without tzinfo should have
        """
        self._alarm = alarm
        self._parent = parent
        self._trigger = trigger
        self._last_ack = acknowledged_until
        self._snooze_until = snoozed_until

    @property
    def acknowledged(self) -> Optional[datetime]:
        """The time in UTC at which this alarm was last acknowledged.

        If the alarm was not acknowledged (dismissed), then this is None.
        """
        ack = self.alarm.ACKNOWLEDGED
        if ack is None:
            return self._last_ack
        if self._last_ack is None:
            return ack
        return max(ack, self._last_ack)

    @property
    def alarm(self) -> Alarm:
        """The alarm component."""
        return self._alarm

    @property
    def parent(self) -> Optional[Parent]:
        """This is the component that contains the alarm.

        This is None if you did not use Alarms.set_component().
        """
        return self._parent

    def is_active(self) -> bool:
        """Whether this alarm is active (True) or acknowledged (False).

        For example, in some calendar software, this is True until the user looks
        at the alarm message and clicked the dismiss button.

        Alarms can be in local time (without a timezone).
        To calculate if the alarm really happened, we need it to be in a timezone.
        If a timezone is required but not given, we throw an IncompleteAlarmInformation.
        """
        acknowledged = self.acknowledged
        if not acknowledged:
            # if nothing is acknowledged, this alarm counts
            return True
        if self._snooze_until is not None and self._snooze_until > acknowledged:
            return True
        trigger = self.trigger
        if trigger.tzinfo is None:
            raise LocalTimezoneMissing(
                "A local timezone is required to check if the alarm is still active. "
                "Use Alarms.set_local_timezone()."
            )
        return trigger > acknowledged

    @property
    def trigger(self) -> date:
        """This is the time to trigger the alarm.

        If the alarm has been snoozed, this can differ from the TRIGGER property.
        """
        if self._snooze_until is not None and self._snooze_until > self._trigger:
            return self._snooze_until
        return self._trigger


class Alarms:
    """Compute the times and states of alarms.

    This is an example using RFC 9074.
    One alarm is 30 minutes before the event and acknowledged.
    Another alarm is 15 minutes before the event and still active.

    >>> from icalendar import Event, Alarms
    >>> event = Event.from_ical(
    ... '''BEGIN:VEVENT
    ... CREATED:20210301T151004Z
    ... UID:AC67C078-CED3-4BF5-9726-832C3749F627
    ... DTSTAMP:20210301T151004Z
    ... DTSTART;TZID=America/New_York:20210302T103000
    ... DTEND;TZID=America/New_York:20210302T113000
    ... SUMMARY:Meeting
    ... BEGIN:VALARM
    ... UID:8297C37D-BA2D-4476-91AE-C1EAA364F8E1
    ... TRIGGER:-PT30M
    ... ACKNOWLEDGED:20210302T150004Z
    ... DESCRIPTION:Event reminder
    ... ACTION:DISPLAY
    ... END:VALARM
    ... BEGIN:VALARM
    ... UID:8297C37D-BA2D-4476-91AE-C1EAA364F8E1
    ... TRIGGER:-PT15M
    ... DESCRIPTION:Event reminder
    ... ACTION:DISPLAY
    ... END:VALARM
    ... END:VEVENT
    ... ''')
    >>> alarms = Alarms(event)
    >>> len(alarms.times)   # all alarms including those acknowledged
    2
    >>> len(alarms.active)  # the alarms that are not acknowledged, yet
    1
    >>> alarms.active[0].trigger  # this alarm triggers 15 minutes before 10:30
    datetime.datetime(2021, 3, 2, 10, 15, tzinfo=ZoneInfo(key='America/New_York'))

    RFC 9074 specifies that alarms can also be triggered by proximity.
    This is not implemented yet.
    """

    def __init__(self, component: Optional[Alarm | Event | Todo] = None):
        """Start computing alarm times."""
        self._absolute_alarms: list[Alarm] = []
        self._start_alarms: list[Alarm] = []
        self._end_alarms: list[Alarm] = []
        self._start: Optional[date] = None
        self._end: Optional[date] = None
        self._parent: Optional[Parent] = None
        self._last_ack: Optional[datetime] = None
        self._snooze_until: Optional[datetime] = None
        self._local_tzinfo: Optional[tzinfo] = None

        if component is not None:
            self.add_component(component)

    def add_component(self, component: Alarm | Parent):
        """Add a component.

        If this is an alarm, it is added.
        Events and Todos are added as a parent and all
        their alarms are added, too.
        """
        if isinstance(component, (Event, Todo)):
            self.set_parent(component)
            self.set_start(component.start)
            self.set_end(component.end)
            if component.is_thunderbird():
                self.acknowledge_until(component.X_MOZ_LASTACK)
                self.snooze_until(component.X_MOZ_SNOOZE_TIME)
            else:
                self.acknowledge_until(component.DTSTAMP)

        for alarm in component.walk("VALARM"):
            self.add_alarm(alarm)

    def set_parent(self, parent: Parent):
        """Set the parent of all the alarms.

        If you would like to collect alarms from a component, use add_component
        """
        if self._parent is not None and self._parent is not parent:
            raise ValueError("You can only set one parent for this alarm calculation.")
        self._parent = parent

    def add_alarm(self, alarm: Alarm) -> None:
        """Optional: Add an alarm component."""
        trigger = alarm.TRIGGER
        if trigger is None:
            return
        if isinstance(trigger, date):
            self._absolute_alarms.append(alarm)
        elif alarm.TRIGGER_RELATED == "START":
            self._start_alarms.append(alarm)
        else:
            self._end_alarms.append(alarm)

    def set_start(self, dt: Optional[date]):
        """Set the start of the component.

        If you have only absolute alarms, this is not required.
        If you have alarms relative to the start of a compoment, set the start here.
        """
        self._start = dt

    def set_end(self, dt: Optional[date]):
        """Set the end of the component.

        If you have only absolute alarms, this is not required.
        If you have alarms relative to the end of a compoment, set the end here.
        """
        self._end = dt

    def _add(self, dt: date, td: timedelta):
        """Add a timedelta to a datetime."""
        if is_date(dt):
            if td.seconds == 0:
                return dt + td
            dt = to_datetime(dt)
        return normalize_pytz(dt + td)

    def acknowledge_until(self, dt: Optional[date]) -> None:
        """This is the time in UTC when all the alarms of this component were acknowledged.

        Only the last call counts.

        Since RFC 9074 (Alarm Extension) was created later,
        calendar implementations differ in how they acknowledge alarms.
        For example, Thunderbird and Google Calendar store the last time
        an event has been acknowledged because of an alarm.
        All alarms that happen before this time count as acknowledged.
        """
        self._last_ack = tzp.localize_utc(dt) if dt is not None else None

    def snooze_until(self, dt: Optional[date]) -> None:
        """This is the time in UTC when all the alarms of this component were snoozed.

        Only the last call counts.

        The alarms are supposed to turn up again at dt when they are not acknowledged
        but snoozed.
        """
        self._snooze_until = tzp.localize_utc(dt) if dt is not None else None

    def set_local_timezone(self, tzinfo: Optional[tzinfo | str]):
        """Set the local timezone.

        Events are sometimes in local time.
        In order to compute the exact time of the alarm, some
        alarms without timezone are considered local.

        Some computations work without setting this, others don't.
        If they need this information, expect a LocalTimezoneMissing exception
        somewhere down the line.
        """
        self._local_tzinfo = tzp.timezone(tzinfo) if isinstance(tzinfo, str) else tzinfo

    @property
    def times(self) -> list[AlarmTime]:
        """Compute and return the times of the alarms given.

        If the information for calculation is incomplete, this will raise a
        IncompleteAlarmInformation exception.

        Please make sure to set all the required parameters before calculating.
        If you forget to set the acknowledged times, that is not problem.
        """
        return (
            self._get_end_alarm_times()
            + self._get_start_alarm_times()
            + self._get_absolute_alarm_times()
        )

    def _repeat(self, first: datetime, alarm: Alarm) -> Generator[datetime]:
        """The times when the alarm is triggered relative to start."""
        yield first  # we trigger at the start
        repeat = alarm.REPEAT
        duration = alarm.DURATION
        if repeat and duration:
            for i in range(1, repeat + 1):
                yield self._add(first, duration * i)

    def _alarm_time(self, alarm: Alarm, trigger: date):
        """Create an alarm time with the additional attributes."""
        if getattr(trigger, "tzinfo", None) is None and self._local_tzinfo is not None:
            trigger = normalize_pytz(trigger.replace(tzinfo=self._local_tzinfo))
        return AlarmTime(
            alarm, trigger, self._last_ack, self._snooze_until, self._parent
        )

    def _get_absolute_alarm_times(self) -> list[AlarmTime]:
        """Return a list of absolute alarm times."""
        return [
            self._alarm_time(alarm, trigger)
            for alarm in self._absolute_alarms
            for trigger in self._repeat(alarm.TRIGGER, alarm)
        ]

    def _get_start_alarm_times(self) -> list[AlarmTime]:
        """Return a list of alarm times relative to the start of the component."""
        if self._start is None and self._start_alarms:
            raise ComponentStartMissing(
                "Use Alarms.set_start because at least one alarm is relative to the start of a component."
            )
        return [
            self._alarm_time(alarm, trigger)
            for alarm in self._start_alarms
            for trigger in self._repeat(self._add(self._start, alarm.TRIGGER), alarm)
        ]

    def _get_end_alarm_times(self) -> list[AlarmTime]:
        """Return a list of alarm times relative to the start of the component."""
        if self._end is None and self._end_alarms:
            raise ComponentEndMissing(
                "Use Alarms.set_end because at least one alarm is relative to the end of a component."
            )
        return [
            self._alarm_time(alarm, trigger)
            for alarm in self._end_alarms
            for trigger in self._repeat(self._add(self._end, alarm.TRIGGER), alarm)
        ]

    @property
    def active(self) -> list[AlarmTime]:
        """The alarm times that are still active and not acknowledged.

        This considers snoozed alarms.

        Alarms can be in local time (without a timezone).
        To calculate if the alarm really happened, we need it to be in a timezone.
        If a timezone is required but not given, we throw an IncompleteAlarmInformation.
        """
        return [alarm_time for alarm_time in self.times if alarm_time.is_active()]


__all__ = [
    "AlarmTime",
    "Alarms",
    "ComponentEndMissing",
    "ComponentStartMissing",
    "IncompleteAlarmInformation",
]



"""Some parameters are specified as default by the RFC 5545.

These tests make sure that these parameter values are not added to the
properties.

Example:

       DTSTART;VALUE=DATE-TIME:20190616T050000Z
equals DTSTART:20190616T050000Z

See https://github.com/collective/icalendar/issues/318
"""

from datetime import datetime

import pytest

from icalendar.cal.event import Event


@pytest.mark.parametrize(
    "attr",
    [
        "DTSTART",
        "DTEND",
        "DTSTAMP",
    ],
)
def test_datetime_in_event(attr):
    """Check that the "VALUE=DATE-TIME" is absent because not needed."""
    event = Event()
    event.add(attr, datetime(2022, 10, 13, 9, 16, 42))
    ics = event.to_ical()
    assert b"VALUE=DATE-TIME" not in ics



from icalendar import vBoolean, vCalAddress
from icalendar.cal.event import Event


def test_vBoolean_can_be_used_as_parameter_issue_500(events):
    """https://github.com/collective/icalendar/issues/500"""
    attendee = vCalAddress("mailto:someone@example.com")
    attendee.params["rsvp"] = vBoolean(True)
    event = Event()
    event.add("attendee", attendee)
    assert event.to_ical() == events.event_with_rsvp.raw_ics



"""This tests if the UID is set be default, closing issue 315.

See https://github.com/collective/icalendar/issues/315
"""

import pytest

from icalendar import Event, Journal, Todo


@pytest.mark.parametrize("cls", [Event, Journal, Todo])
def test_uid_is_included(cls):
    """Make sure we have a UID!"""
    component = cls.new()
    assert "uid" in component
    assert component.uid != ""
    assert "UID" in component.to_ical().decode()



"""Test the conversion of components."""

import json
from icalendar import Calendar, Event, Todo, Journal
import pytest

@pytest.mark.parametrize(
    ["component_class"],
    [
        (Calendar,),
        (Event,),
        (Todo,),
        (Journal,),
    ]
)
def test_component_conversion(component_class):
    """test conversion of components to jCal."""
    component = component_class()
    jcal = component.to_jcal()
    assert len(jcal) == 3
    assert jcal[0] == component_class.name.lower()
    assert jcal[1] == []
    assert jcal[2] == []


@pytest.mark.parametrize(
    ["component_class"],
    [
        (Event,),
        (Todo,),
        (Journal,),
    ]
)
def test_nesting(component_class):
    """Nested components turn up in an sub array."""
    calendar = Calendar()
    calendar.add_component(component_class())
    jcal = calendar.to_jcal()
    assert len(jcal) == 3
    assert jcal[0] == "vcalendar"
    assert jcal[1] == []
    assert len(jcal[2]) == 1
    assert jcal[2][0] == [
        component_class.name.lower(),
        [],
        [],
    ]

def test_apply_json_serialization():
    """Check that we can convert to JSON."""
    calendar = Calendar()
    calendar.add_component(Event())
    assert calendar.to_json() == json.dumps(calendar.to_jcal())


# tests/test_cli_stdin_argparse.py
import io
import os
import sys
import unittest
import tempfile
import contextlib
from datetime import datetime
from unittest import mock

from icalendar import Calendar, cli
from icalendar.compatibility import ZoneInfo


INPUT = """
BEGIN:VCALENDAR
VERSION:2.0
CALSCALE:GREGORIAN
BEGIN:VEVENT
SUMMARY:Test Summary
ORGANIZER:organizer@test.test
ATTENDEE:attendee1@example.com
ATTENDEE:attendee2@test.test
COMMENT:Comment
DTSTART;TZID=Europe/Warsaw:20220820T103400
DTEND;TZID=Europe/Warsaw:20220820T113400
LOCATION:New Amsterdam, 1000 Sunrise Test Street
DESCRIPTION: Test Description
END:VEVENT
BEGIN:VEVENT
ORGANIZER:organizer@test.test
ATTENDEE:attendee1@example.com
SUMMARY:Test summary
DTSTART;TZID=Europe/Warsaw:20220820T200000
DTEND;TZID=Europe/Warsaw:20220820T203000
LOCATION:New Amsterdam, 1010 Test Street
DESCRIPTION:Test Description\\nThis one is multiline
END:VEVENT
BEGIN:VEVENT
UID:1
SUMMARY:TEST
DTSTART:20220511
DURATION:P5D
END:VEVENT
END:VCALENDAR
""".lstrip()


def _local_datetime(dt: str) -> str:
    """
    Match the CLI's behavior: display datetimes in the *local* timezone.
    The source times are Europe/Warsaw; convert to local tz for comparison.
    """
    return (
        datetime.strptime(dt, "%Y%m%dT%H%M%S")
        .replace(tzinfo=ZoneInfo("Europe/Warsaw"))
        .astimezone()
        .strftime("%c")
    )


# datetimes are displayed in the local timezone, so build expected strings at runtime
_firststart = _local_datetime("20220820T103400")
_firstend = _local_datetime("20220820T113400")
_secondstart = _local_datetime("20220820T200000")
_secondend = _local_datetime("20220820T203000")

PROPER_OUTPUT = f"""    Organizer: organizer <organizer@test.test>
    Attendees:
     attendee1 <attendee1@example.com>
     attendee2 <attendee2@test.test>
    Summary    : Test Summary
    Starts     : {_firststart}
    End        : {_firstend}
    Duration   : 1:00:00
    Location   : New Amsterdam, 1000 Sunrise Test Street
    Comment    : Comment
    Description:
      Test Description

    Organizer: organizer <organizer@test.test>
    Attendees:
     attendee1 <attendee1@example.com>
    Summary    : Test summary
    Starts     : {_secondstart}
    End        : {_secondend}
    Duration   : 0:30:00
    Location   : New Amsterdam, 1010 Test Street
    Comment    : 
    Description:
     Test Description
     This one is multiline

    Organizer: 
    Attendees:

    Summary    : TEST
    Starts     : Wed May 11 00:00:00 2022
    End        : Mon May 16 00:00:00 2022
    Duration   : 5 days, 0:00:00
    Location   : 
    Comment    : 
    Description:
     

""".replace("\r\n", "\n")  # normalize newlines just in case


class TestIcalendarCLIArgparseStdin(unittest.TestCase):
    def test_cli_with_stdin_dash(self):
        """
        Validate that `type=argparse.FileType("r", ...)` maps '-' to sys.stdin
        and the CLI prints the expected, fully formatted output.
        """
        # Provide INPUT via stdin
        fake_stdin = io.StringIO(INPUT)
        captured = io.StringIO()

        with mock.patch.object(sys, "stdin", fake_stdin), \
             mock.patch.object(sys, "argv", ["icalendar", "-"]), \
             contextlib.redirect_stdout(captured):
            cli.main()

        self.assertEqual(captured.getvalue(), PROPER_OUTPUT)

    def test_cli_with_file_path(self):
        """
        Validate that passing a real file path also yields the same output.
        """
        with tempfile.NamedTemporaryFile("w", delete=False, encoding="utf-8") as tf:
            tf.write(INPUT)
            tmpname = tf.name

        try:
            captured = io.StringIO()
            with mock.patch.object(sys, "argv", ["icalendar", tmpname]), \
                 contextlib.redirect_stdout(captured):
                cli.main()
            self.assertEqual(captured.getvalue(), PROPER_OUTPUT)
        finally:
            os.unlink(tmpname)

    def test_view_direct_matches_cli(self):
        """
        Optional: sanity check that the lower-level view(event) path
        still matches what the CLI emits, to guard against formatting drift.
        """
        cal = Calendar.from_ical(INPUT)
        manual = []
        for evt in cal.walk("vevent"):
            manual.append(cli.view(evt))
            manual.append("")  # the CLI prints a blank line after each event
        manual_output = "\n".join(manual) + "\n"  # trailing newline

        self.assertEqual(manual_output, PROPER_OUTPUT)


if __name__ == "__main__":
    unittest.main()



import itertools
import re
from datetime import datetime, timedelta

import pytest

import icalendar
from icalendar import prop
from icalendar.cal.calendar import Calendar
from icalendar.cal.component import Component
from icalendar.cal.event import Event
from icalendar.prop import tzid_from_dt


def test_cal_Component(calendar_component):
    """A component is like a dictionary with extra methods and attributes."""
    assert calendar_component
    assert calendar_component.is_empty()


def test_nonempty_calendar_component(calendar_component):
    """Every key defines a property.A property can consist of either a
    single item. This can be set with a single value...
    """
    calendar_component["prodid"] = "-//max m//icalendar.mxm.dk/"
    assert not calendar_component.is_empty()
    assert calendar_component == Calendar({"PRODID": "-//max m//icalendar.mxm.dk/"})

    # or with a list
    calendar_component["ATTENDEE"] = ["Max M", "Rasmussen"]
    assert calendar_component == Calendar(
        {"ATTENDEE": ["Max M", "Rasmussen"], "PRODID": "-//max m//icalendar.mxm.dk/"}
    )


def test_add_multiple_values(event_component):
    """add multiple values to a property.

    If you use the add method you don't have to considder if a value is
    a list or not.
    """
    # add multiple values at once
    event_component.add("attendee", ["test@test.com", "test2@test.com"])

    # or add one per line
    event_component.add("attendee", "maxm@mxm.dk")
    event_component.add("attendee", "test@example.dk")

    # add again multiple values at once to very concatenaton of lists
    event_component.add("attendee", ["test3@test.com", "test4@test.com"])

    assert event_component == Event(
        {
            "ATTENDEE": [
                prop.vCalAddress("test@test.com"),
                prop.vCalAddress("test2@test.com"),
                prop.vCalAddress("maxm@mxm.dk"),
                prop.vCalAddress("test@example.dk"),
                prop.vCalAddress("test3@test.com"),
                prop.vCalAddress("test4@test.com"),
            ]
        }
    )


def test_get_content_directly(c):
    """You can get the values back directly ..."""
    c.add("prodid", "-//my product//")
    assert c["prodid"] == prop.vText("-//my product//")
    # ... or decoded to a python type
    assert c.decoded("prodid") == b"-//my product//"


def test_get_default_value(c):
    """With default values for non existing properties"""
    assert c.decoded("version", "No Version") == "No Version"


def test_default_list_example(c):
    c.add("rdate", [datetime(2013, 3, 28), datetime(2013, 3, 27)])
    assert isinstance(c.decoded("rdate"), prop.vDDDLists)


def test_render_component(calendar_component):
    """The component can render itself in the RFC 5545 format."""
    calendar_component.add("attendee", "Max M")
    assert (
        calendar_component.to_ical()
        == b"BEGIN:VCALENDAR\r\nATTENDEE:Max M\r\nEND:VCALENDAR\r\n"
    )


def test_nested_component_event_ics(filled_event_component):
    """Check the ical string of the event component."""
    assert filled_event_component.to_ical() == (
        b"BEGIN:VEVENT\r\nDTEND:20000102T000000\r\n"
        + b"DTSTART:20000101T000000\r\nSUMMARY:A brief history of time\r"
        + b"\nEND:VEVENT\r\n"
    )


def test_nested_components(calendar_component, filled_event_component):
    """Components can be nested, so You can add a subcomponent. Eg a calendar
    holds events."""
    assert calendar_component.subcomponents == [
        Event(
            {
                "DTEND": "20000102T000000",
                "DTSTART": "20000101T000000",
                "SUMMARY": "A brief history of time",
            }
        )
    ]


def test_walk_filled_calendar_component(calendar_component, filled_event_component):
    """We can walk over nested componentes with the walk method."""
    assert [i.name for i in calendar_component.walk()] == ["VCALENDAR", "VEVENT"]


def test_filter_walk(calendar_component, filled_event_component):
    """We can also just walk over specific component types, by filtering
    them on their name."""
    assert [i.name for i in calendar_component.walk("VEVENT")] == ["VEVENT"]
    assert [i["dtstart"] for i in calendar_component.walk("VEVENT")] == [
        "20000101T000000"
    ]


def test_recursive_property_items(calendar_component, filled_event_component):
    """We can enumerate property items recursively with the property_items
    method."""
    calendar_component.add("attendee", "Max M")
    assert calendar_component.property_items() == [
        ("BEGIN", b"VCALENDAR"),
        ("ATTENDEE", prop.vCalAddress("Max M")),
        ("BEGIN", b"VEVENT"),
        ("DTEND", "20000102T000000"),
        ("DTSTART", "20000101T000000"),
        ("SUMMARY", "A brief history of time"),
        ("END", b"VEVENT"),
        ("END", b"VCALENDAR"),
    ]


def test_flat_property_items(calendar_component, filled_event_component):
    """We can also enumerate property items just under the component."""
    assert calendar_component.property_items(recursive=False) == [
        ("BEGIN", b"VCALENDAR"),
        ("ATTENDEE", prop.vCalAddress("Max M")),
        ("END", b"VCALENDAR"),
    ]


def test_flat_property_items(filled_event_component):
    """Flat enumeration on the event."""
    assert filled_event_component.property_items(recursive=False) == [
        ("BEGIN", b"VEVENT"),
        ("DTEND", "20000102T000000"),
        ("DTSTART", "20000101T000000"),
        ("SUMMARY", "A brief history of time"),
        ("END", b"VEVENT"),
    ]


def test_indent():
    """Text fields which span multiple mulitple lines require proper indenting"""
    c = Calendar()
    c["description"] = "Paragraph one\n\nParagraph two"
    assert c.to_ical() == (
        b"BEGIN:VCALENDAR\r\nDESCRIPTION:Paragraph one\\n\\nParagraph two"
        + b"\r\nEND:VCALENDAR\r\n"
    )


def test_INLINE_properties(calendar_with_resources):
    """INLINE properties have their values on one property line. Note the
    double quoting of the value with a colon in it.
    """
    assert calendar_with_resources == Calendar(
        {"RESOURCES": 'Chair, Table, "Room: 42"'}
    )
    assert calendar_with_resources.to_ical() == (
        b'BEGIN:VCALENDAR\r\nRESOURCES:Chair\\, Table\\, "Room: 42"\r\n'
        + b"END:VCALENDAR\r\n"
    )


def test_get_inline(calendar_with_resources):
    """The inline values must be handled by the get_inline() and
    set_inline() methods.
    """
    assert calendar_with_resources.get_inline("resources", decode=0) == [
        "Chair",
        "Table",
        "Room: 42",
    ]


def test_get_inline_decoded(calendar_with_resources):
    """These can also be decoded"""
    assert calendar_with_resources.get_inline("resources", decode=1) == [
        b"Chair",
        b"Table",
        b"Room: 42",
    ]


def test_set_inline(calendar_with_resources):
    """You can set them directly ..."""
    calendar_with_resources.set_inline(
        "resources", ["A", "List", "of", "some, recources"], encode=1
    )
    assert calendar_with_resources["resources"] == 'A,List,of,"some, recources"'
    assert calendar_with_resources.get_inline("resources", decode=0) == [
        "A",
        "List",
        "of",
        "some, recources",
    ]


def test_inline_free_busy_inline(c):
    c["freebusy"] = (
        "19970308T160000Z/PT3H,19970308T200000Z/PT1H,"
        + "19970308T230000Z/19970309T000000Z"
    )
    assert c.get_inline("freebusy", decode=0) == [
        "19970308T160000Z/PT3H",
        "19970308T200000Z/PT1H",
        "19970308T230000Z/19970309T000000Z",
    ]

    freebusy = c.get_inline("freebusy", decode=1)
    assert isinstance(freebusy[0][0], datetime)
    assert isinstance(freebusy[0][1], timedelta)


def test_cal_Component_add(comp, tzp):
    """Test the for timezone correctness: dtstart should preserve it's
    timezone, created, dtstamp and last-modified must be in UTC.
    """
    comp.add("dtstart", tzp.localize(datetime(2010, 10, 10, 10, 0, 0), "Europe/Vienna"))
    comp.add("created", datetime(2010, 10, 10, 12, 0, 0))
    comp.add("dtstamp", tzp.localize(datetime(2010, 10, 10, 14, 0, 0), "Europe/Vienna"))
    comp.add("last-modified", tzp.localize_utc(datetime(2010, 10, 10, 16, 0, 0)))

    lines = comp.to_ical().splitlines()
    assert b"DTSTART;TZID=Europe/Vienna:20101010T100000" in lines
    assert b"CREATED:20101010T120000Z" in lines
    assert b"DTSTAMP:20101010T120000Z" in lines
    assert b"LAST-MODIFIED:20101010T160000Z" in lines


def test_cal_Component_add_no_reencode(comp):
    """Already encoded values should not be re-encoded."""
    comp.add("ATTACH", "me")
    comp.add("ATTACH", "you", encode=False)
    binary = prop.vBinary("us")
    comp.add("ATTACH", binary)

    assert comp["ATTACH"] == ["me", "you", binary]


def test_cal_Component_add_property_parameter(comp):
    """Test the for timezone correctness: dtstart should preserve it's
    timezone, crated, dtstamp and last-modified must be in UTC.
    """
    comp.add("X-TEST-PROP", "tryout.", parameters={"prop1": "val1", "prop2": "val2"})
    lines = comp.to_ical().splitlines()
    assert b"X-TEST-PROP;PROP1=val1;PROP2=val2:tryout." in lines


comp_prop = pytest.mark.parametrize(
    ("component_name", "property_name"),
    [
        ("VEVENT", "DTSTART"),
        ("VEVENT", "DTEND"),
        ("VEVENT", "RECURRENCE-ID"),
        ("VTODO", "DUE"),
    ],
)


@comp_prop
def test_cal_Component_from_ical(component_name, property_name, tzp):
    """Check for proper handling of TZID parameter of datetime properties"""
    component_str = "BEGIN:" + component_name + "\n"
    component_str += property_name + ";TZID=America/Denver:"
    component_str += "20120404T073000\nEND:" + component_name
    component = Component.from_ical(component_str)
    assert tzid_from_dt(component[property_name].dt) == "America/Denver"


@comp_prop
def test_cal_Component_from_ical_2(component_name, property_name, tzp):
    """Check for proper handling of TZID parameter of datetime properties"""
    component_str = "BEGIN:" + component_name + "\n"
    component_str += property_name + ":"
    component_str += "20120404T073000\nEND:" + component_name
    component = Component.from_ical(component_str)
    assert component[property_name].dt.tzinfo is None


def test_cal_Component_to_ical_property_order():
    component_str = [
        b"BEGIN:VEVENT",
        b"DTSTART:19970714T170000Z",
        b"DTEND:19970715T035959Z",
        b"SUMMARY:Bastille Day Party",
        b"END:VEVENT",
    ]
    component = Component.from_ical(b"\r\n".join(component_str))

    sorted_str = component.to_ical().splitlines()
    assert sorted_str != component_str
    assert set(sorted_str) == set(component_str)

    preserved_str = component.to_ical(sorted=False).splitlines()
    assert preserved_str == component_str


def test_cal_Component_to_ical_parameter_order():
    component_str = [
        b"BEGIN:VEVENT",
        b"X-FOOBAR;C=one;A=two;B=three:helloworld.",
        b"END:VEVENT",
    ]
    component = Component.from_ical(b"\r\n".join(component_str))

    sorted_str = component.to_ical().splitlines()
    assert sorted_str[0] == component_str[0]
    assert sorted_str[1] == b"X-FOOBAR;A=two;B=three;C=one:helloworld."
    assert sorted_str[2] == component_str[2]

    preserved_str = component.to_ical(sorted=False).splitlines()
    assert preserved_str == component_str


@pytest.fixture
def repr_example(c):
    class ReprExample:
        component = c
        component["key1"] = "value1"
        calendar = Calendar()
        calendar["key1"] = "value1"
        event = Event()
        event["key1"] = "value1"
        nested = Component(key1="VALUE1")
        nested.add_component(component)
        nested.add_component(calendar)

    return ReprExample


def test_repr_component(repr_example):
    """Test correct class representation."""
    assert re.match(r"Component\({u?'KEY1': u?'value1'}\)", str(repr_example.component))


def test_repr_calendar(repr_example):
    assert re.match(r"VCALENDAR\({u?'KEY1': u?'value1'}\)", str(repr_example.calendar))


def test_repr_event(repr_example):
    assert re.match(r"VEVENT\({u?'KEY1': u?'value1'}\)", str(repr_example.event))


def test_nested_components(repr_example):
    """Representation of nested Components"""
    repr_example.calendar.add_component(repr_example.event)
    print(repr_example.nested)
    assert re.match(
        r"Component\({u?'KEY1': u?'VALUE1'}, "
        r"Component\({u?'KEY1': u?'value1'}\), "
        r"VCALENDAR\({u?'KEY1': u?'value1'}, "
        r"VEVENT\({u?'KEY1': u?'value1'}\)\)\)",
        str(repr_example.nested),
    )


def test_component_factory_VEVENT(factory):
    """Check the events in the component factory"""
    component = factory["VEVENT"]
    event = component(dtstart="19700101")
    assert event.to_ical() == b"BEGIN:VEVENT\r\nDTSTART:19700101\r\nEND:VEVENT\r\n"


def test_component_factory_VCALENDAR(factory):
    """Check the VCALENDAR in the factory."""
    assert factory.get("VCALENDAR") == icalendar.cal.Calendar


def test_minimal_calendar_component_with_one_event():
    """Setting up a minimal calendar component looks like this"""
    cal = Calendar()

    # Some properties are required to be compliant
    cal["prodid"] = "-//My calendar product//mxm.dk//"
    cal["version"] = "2.0"

    # We also need at least one subcomponent for a calendar to be compliant
    event = Event()
    event["summary"] = "Python meeting about calendaring"
    event["uid"] = "42"
    event.add("dtstart", datetime(2005, 4, 4, 8, 0, 0))
    cal.add_component(event)
    assert (
        cal.subcomponents[0].to_ical()
        == b"BEGIN:VEVENT\r\nSUMMARY:Python meeting about calendaring\r\n"
        + b"DTSTART:20050404T080000\r\nUID:42\r\n"
        + b"END:VEVENT\r\n"
    )


def test_calendar_with_parsing_errors_includes_all_events(calendars):
    """Parsing a complete calendar from a string will silently ignore wrong
    events but adding the error information to the component's 'errors'
    attribute. The error in the following is the third EXDATE: it has an
    empty DATE.
    """
    event_descriptions = [
        e["DESCRIPTION"].to_ical() for e in calendars.parsing_error.walk("VEVENT")
    ]
    assert event_descriptions == [b"Perfectly OK event", b"Wrong event"]


def test_calendar_with_parsing_errors_has_an_error_in_one_event(calendars):
    """Parsing a complete calendar from a string will silently ignore wrong
    events but adding the error information to the component's 'errors'
    attribute. The error in the following is the third EXDATE: it has an
    empty DATE.
    """
    errors = [e.errors for e in calendars.parsing_error.walk("VEVENT")]
    assert errors == [[], [("EXDATE", "Expected datetime, date, or time. Got: ''")]]


def test_cal_strict_parsing(calendars):
    """If components are damaged, we raise an exception."""
    with pytest.raises(ValueError):
        calendars.parsing_error_in_UTC_offset


def test_cal_ignore_errors_parsing(calendars, vUTCOffset_ignore_exceptions):
    """If we diable the errors, we should be able to put the calendar back together."""
    assert (
        calendars.parsing_error_in_UTC_offset.to_ical()
        == calendars.parsing_error_in_UTC_offset.raw_ics
    )


@pytest.mark.parametrize(
    ("calendar", "other_calendar"),
    itertools.product(
        [
            "issue_156_RDATE_with_PERIOD_TZID_khal",
            "issue_156_RDATE_with_PERIOD_TZID_khal_2",
            "issue_178_custom_component_contains_other",
            "issue_178_custom_component_inside_other",
            "issue_526_calendar_with_events",
            "issue_526_calendar_with_different_events",
            "issue_526_calendar_with_event_subset",
        ],
        repeat=2,
    ),
)
def test_comparing_calendars(calendars, calendar, other_calendar, tzp):
    are_calendars_equal = calendars[calendar] == calendars[other_calendar]
    are_calendars_actually_equal = calendar == other_calendar
    assert are_calendars_equal == are_calendars_actually_equal


@pytest.mark.parametrize(
    ("calendar", "shuffeled_calendar"),
    [
        (
            "issue_526_calendar_with_events",
            "issue_526_calendar_with_shuffeled_events",
        ),
    ],
)
def test_calendars_with_same_subcomponents_in_different_order_are_equal(
    calendars, calendar, shuffeled_calendar
):
    assert (
        calendars[calendar].subcomponents != calendars[shuffeled_calendar].subcomponents
    )
    assert calendars[calendar] == calendars[shuffeled_calendar]



"""Converting from jCal into iCalendar - :rfc:`7265`

- parameters
- properties
- components

"""

import pytest

from icalendar.parser import Parameters


@pytest.mark.parametrize("jcal", [{}, {"value": "DATE-TIME"}, {"x-slack": "30.3"}])
def test_parameters_from_jcal(jcal):
    """General constrints for parameter parsing.

    > When converting [...] property parameter names,
    > the names SHOULD be converted to uppercase.
    """
    params = Parameters.from_jcal(jcal)
    assert len(params) == len(jcal)
    assert all(key.isupper() for key in params)



"""Parameter access for icalendar.

Related:

- :rfc:`5545`, Section 3.2. Property Parameters
- :rfc:`7986`, Section 6. Property Parameters
- :rfc:`9253`
- https://github.com/collective/icalendar/issues/798
"""

from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Callable, Optional, TypeVar

from icalendar import enums

if TYPE_CHECKING:
    from datetime import timedelta
    from enum import Enum


if TYPE_CHECKING:
    from icalendar.prop import VPROPERTY


def _default_return_none() -> Optional[str]:
    """Return None by default."""
    return None


def _default_return_string() -> str:
    """Return None by default."""
    return ""


T = TypeVar("T")


def string_parameter(
    name: str,
    doc: str,
    default: Callable = _default_return_none,
    convert: Optional[Callable[[str], T]] = None,
    convert_to: Optional[Callable[[T], str]] = None,
) -> property:
    """Return a parameter with a quoted value (case sensitive)."""

    if convert_to is None:
        convert_to = convert

    @functools.wraps(default)
    def fget(self: VPROPERTY) -> Optional[str]:
        value = self.params.get(name)
        if value is None:
            return default()
        return convert(value) if convert else value

    def fset(self: VPROPERTY, value: str | None):
        if value is None:
            fdel(self)
        else:
            self.params[name] = convert_to(value) if convert_to else value

    def fdel(self: VPROPERTY):
        self.params.pop(name, None)

    return property(fget, fset, fdel, doc=doc)


ALTREP = string_parameter(
    "ALTREP",
    """ALTREP - Specify an alternate text representation for the property value.

Description:
    This parameter specifies a URI that points to an
    alternate representation for a textual property value.  A property
    specifying this parameter MUST also include a value that reflects
    the default representation of the text value.  The URI parameter
    value MUST be specified in a quoted-string.

.. note::

    While there is no restriction imposed on the URI schemes
    allowed for this parameter, Content Identifier (CID) :rfc:`2392`,
    HTTP :rfc:`2616`, and HTTPS :rfc:`2818` are the URI schemes most
    commonly used by current implementations.
""",
)

CN = string_parameter(
    "CN",
    """Specify the common name to be associated with the calendar user specified.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter specifies the common name
    to be associated with the calendar user specified by the property.
    The parameter value is text.  The parameter value can be used for
    display text to be associated with the calendar address specified
    by the property.
""",
    default=_default_return_string,
)


def _default_return_individual() -> enums.CUTYPE | str:
    """Default value."""
    return enums.CUTYPE.INDIVIDUAL


def _convert_enum(enum: type[Enum]) -> Callable[[str], Enum]:
    def convert(value: str) -> str:
        """Convert if possible."""
        try:
            return enum(value.upper())
        except ValueError:
            return value

    return convert


CUTYPE = string_parameter(
    "CUTYPE",
    """Identify the type of calendar user specified by the property.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter identifies the type of
    calendar user specified by the property.  If not specified on a
    property that allows this parameter, the default is INDIVIDUAL.
    Applications MUST treat x-name and iana-token values they don't
    recognize the same way as they would the UNKNOWN value.
""",
    default=_default_return_individual,
    convert=_convert_enum(enums.CUTYPE),
)


def quoted_list_parameter(name: str, doc: str) -> property:
    """Return a parameter that contains a quoted list."""

    def fget(self: VPROPERTY) -> tuple[str]:
        value = self.params.get(name)
        if value is None:
            return ()
        if isinstance(value, str):
            return tuple(value.split(","))
        return value

    def fset(self: VPROPERTY, value: str | tuple[str]):
        if value == ():
            fdel(self)
        else:
            self.params[name] = (value,) if isinstance(value, str) else value

    def fdel(self: VPROPERTY):
        self.params.pop(name, None)

    return property(fget, fset, fdel, doc=doc)


DELEGATED_FROM = quoted_list_parameter(
    "DELEGATED-FROM",
    """Specify the calendar users that have delegated their participation to the calendar user specified by the property.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  This parameter specifies those calendar
    users that have delegated their participation in a group-scheduled
    event or to-do to the calendar user specified by the property.
    The individual calendar address parameter values MUST each be
    specified in a quoted-string.
""",  # noqa: E501
)

DELEGATED_TO = quoted_list_parameter(
    "DELEGATED-TO",
    """Specify the calendar users to whom the calendar user specified by the property has delegated participation.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  This parameter specifies those calendar
    users whom have been delegated participation in a group-scheduled
    event or to-do by the calendar user specified by the property.
    The individual calendar address parameter values MUST each be
    specified in a quoted-string.
    """,  # noqa: E501
)

DIR = string_parameter(
    "DIR",
    """Specify reference to a directory entry associated with the calendar user specified by the property.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter specifies a reference to
    the directory entry associated with the calendar user specified by
    the property.  The parameter value is a URI.  The URI parameter
    value MUST be specified in a quoted-string.

.. note::

    While there is no restriction imposed on the URI schemes
    allowed for this parameter, CID :rfc:`2392`, DATA :rfc:`2397`, FILE
    :rfc:`1738`, FTP :rfc:`1738`, HTTP :rfc:`2616`, HTTPS :rfc:`2818`, LDAP
    :rfc:`4516`, and MID :rfc:`2392` are the URI schemes most commonly
    used by current implementations.
""",  # noqa: E501
)


def _default_return_busy() -> enums.FBTYPE | str:
    """Default value."""
    return enums.FBTYPE.BUSY


FBTYPE = string_parameter(
    "FBTYPE",
    """Specify the free or busy time type.

Description:
    This parameter specifies the free or busy time type.
    The value FREE indicates that the time interval is free for
    scheduling.  The value BUSY indicates that the time interval is
    busy because one or more events have been scheduled for that
    interval.  The value BUSY-UNAVAILABLE indicates that the time
    interval is busy and that the interval can not be scheduled.  The
    value BUSY-TENTATIVE indicates that the time interval is busy
    because one or more events have been tentatively scheduled for
    that interval.  If not specified on a property that allows this
    parameter, the default is BUSY.  Applications MUST treat x-name
    and iana-token values they don't recognize the same way as they
    would the BUSY value.
""",
    default=_default_return_busy,
    convert=_convert_enum(enums.FBTYPE),
)

LANGUAGE = string_parameter(
    "LANGUAGE",
    """Specify the language for text values in a property or property parameter.

Description:
    This parameter identifies the language of the text in
    the property value and of all property parameter values of the
    property.  The value of the "LANGUAGE" property parameter is that
    defined in :rfc:`5646`.

    For transport in a MIME entity, the Content-Language header field
    can be used to set the default language for the entire body part.
    Otherwise, no default language is assumed.
""",
)

MEMBER = quoted_list_parameter(
    "MEMBER",
    """Specify the group or list membership of the calendar user specified by the property.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter identifies the groups or
    list membership for the calendar user specified by the property.
    The parameter value is either a single calendar address in a
    quoted-string or a COMMA-separated list of calendar addresses,
    each in a quoted-string.  The individual calendar address
    parameter values MUST each be specified in a quoted-string.
""",  # noqa: E501
)


def _default_return_needs_action() -> enums.PARTSTAT | str:
    """Default value."""
    return enums.PARTSTAT.NEEDS_ACTION


PARTSTAT = string_parameter(
    "PARTSTAT",
    """Specify the participation status for the calendar user specified by the property.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter identifies the
    participation status for the calendar user specified by the
    property value.  The parameter values differ depending on whether
    they are associated with a group-scheduled "VEVENT", "VTODO", or
    "VJOURNAL".  The values MUST match one of the values allowed for
    the given calendar component.  If not specified on a property that
    allows this parameter, the default value is NEEDS-ACTION.
    Applications MUST treat x-name and iana-token values they don't
    recognize the same way as they would the NEEDS-ACTION value.
""",
    default=_default_return_needs_action,
    convert=_convert_enum(enums.PARTSTAT),
)


def _default_range_none() -> Optional[enums.RANGE | str]:
    return None


RANGE = string_parameter(
    "RANGE",
    """Specify the effective range of recurrence instances from the instance specified by the recurrence identifier specified by the property.

Description:
    This parameter can be specified on a property that
    specifies a recurrence identifier.  The parameter specifies the
    effective range of recurrence instances that is specified by the
    property.  The effective range is from the recurrence identifier
    specified by the property.  If this parameter is not specified on
    an allowed property, then the default range is the single instance
    specified by the recurrence identifier value of the property.  The
    parameter value can only be "THISANDFUTURE" to indicate a range
    defined by the recurrence identifier and all subsequent instances.
    The value "THISANDPRIOR" is deprecated by this revision of
    iCalendar and MUST NOT be generated by applications.
""",  # noqa: E501
    default=_default_range_none,
    convert=_convert_enum(enums.RANGE),
)


def _default_related() -> enums.RELATED | str:
    return enums.RELATED.START


RELATED = string_parameter(
    "RELATED",
    """Specify the relationship of the alarm trigger with respect to the start or end of the calendar component.

Description:
    This parameter can be specified on properties that
    specify an alarm trigger with a "DURATION" value type.  The
    parameter specifies whether the alarm will trigger relative to the
    start or end of the calendar component.  The parameter value START
    will set the alarm to trigger off the start of the calendar
    component; the parameter value END will set the alarm to trigger
    off the end of the calendar component.  If the parameter is not
    specified on an allowable property, then the default is START.

""",  # noqa: E501
    default=_default_related,
    convert=_convert_enum(enums.RELATED),
)


def _default_req_participant() -> enums.ROLE | str:
    return enums.ROLE.REQ_PARTICIPANT


ROLE = string_parameter(
    "ROLE",
    """Specify the participation role for the calendar user specified by the property.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter specifies the participation
    role for the calendar user specified by the property in the group
    schedule calendar component.  If not specified on a property that
    allows this parameter, the default value is REQ-PARTICIPANT.
    Applications MUST treat x-name and iana-token values they don't
    recognize the same way as they would the REQ-PARTICIPANT value.
""",
    default=_default_req_participant,
    convert=_convert_enum(enums.ROLE),
)


def boolean_parameter(name: str, default: bool, doc: str) -> property:  # noqa: FBT001
    def _default() -> bool:
        return default

    return string_parameter(
        name,
        doc,
        default=_default,
        convert=lambda x: x.upper() == "TRUE",
        convert_to=lambda x: "TRUE" if x else "FALSE",
    )


RSVP = boolean_parameter(
    "RSVP",
    False,  # noqa: FBT003
    """Specify whether there is an expectation of a favor of anreply from the calendar user specified by the property value.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter identifies the expectation
    of a reply from the calendar user specified by the property value.
    This parameter is used by the "Organizer" to request a
    participation status reply from an "Attendee" of a group-scheduled
    event or to-do.  If not specified on a property that allows this
    parameter, the default value is ``False``.
""",  # noqa: E501
)

SENT_BY = string_parameter(
    "SENT-BY",
    """Specify the calendar user that is acting on behalf of the calendar user specified by the property.

Description:
    This parameter can be specified on properties with a
    CAL-ADDRESS value type.  The parameter specifies the calendar user
    that is acting on behalf of the calendar user specified by the
    property.  The parameter value MUST be a mailto URI as defined in
    :rfc:`2368`.  The individual calendar address parameter values MUST
    each be specified in a quoted-string.
""",  # noqa: E501
)

TZID = string_parameter(
    "TZID",
    """Specify the identifier for the time zone definition for a time component in the property value.

Description:
    This parameter MUST be specified on the "DTSTART",
    "DTEND", "DUE", "EXDATE", and "RDATE" properties when either a
    DATE-TIME or TIME value type is specified and when the value is
    neither a UTC or a "floating" time.  Refer to the DATE-TIME or
    TIME value type definition for a description of UTC and "floating
    time" formats.  This property parameter specifies a text value
    that uniquely identifies the "VTIMEZONE" calendar component to be
    used when evaluating the time portion of the property.  The value
    of the "TZID" property parameter will be equal to the value of the
    "TZID" property for the matching time zone definition.  An
    individual "VTIMEZONE" calendar component MUST be specified for
    each unique "TZID" parameter value specified in the iCalendar
    object.

    The parameter MUST be specified on properties with a DATE-TIME
    value if the DATE-TIME is not either a UTC or a "floating" time.
    Failure to include and follow VTIMEZONE definitions in iCalendar
    objects may lead to inconsistent understanding of the local time
    at any given location.

    The presence of the SOLIDUS character as a prefix, indicates that
    this "TZID" represents a unique ID in a globally defined time zone
    registry (when such registry is defined).

.. note::

    This document does not define a naming convention for
    time zone identifiers.  Implementers may want to use the naming
    conventions defined in existing time zone specifications such
    as the public-domain TZ database (TZDB). The specification of
    globally unique time zone identifiers is not addressed by this
    document and is left for future study.
""",  # noqa: E501
)


def _default_return_parent() -> enums.RELTYPE:
    return enums.RELTYPE.PARENT


RELTYPE = string_parameter(
    "RELTYPE",
    """Specify the type of hierarchical relationship associated with a component.

Conformance:
    :rfc:`5545` introduces the RELTYPE property parameter.
    :rfc:`9253` adds new values.

Description:
    This parameter can be specified on a property that
    references another related calendar.  The parameter specifies the
    hierarchical relationship type of the calendar component
    referenced by the property.  The parameter value can be PARENT, to
    indicate that the referenced calendar component is a superior of
    calendar component; CHILD to indicate that the referenced calendar
    component is a subordinate of the calendar component; or SIBLING
    to indicate that the referenced calendar component is a peer of
    the calendar component.  If this parameter is not specified on an
    allowable property, the default relationship type is PARENT.
    Applications MUST treat x-name and iana-token values they don't
    recognize the same way as they would the PARENT value.
""",
    default=_default_return_parent,
    convert=_convert_enum(enums.RELTYPE),
)


def _get_value(self: VPROPERTY) -> str:
    """The VALUE parameter or the default.

    Purpose:
        VALUE explicitly specify the value type format for a property value.

    Description:
        This parameter specifies the value type and format of
        the property value.  The property values MUST be of a single value
        type.  For example, a "RDATE" property cannot have a combination
        of DATE-TIME and TIME value types.

        If the property's value is the default value type, then this
        parameter need not be specified.  However, if the property's
        default value type is overridden by some other allowable value
        type, then this parameter MUST be specified.

        Applications MUST preserve the value data for ``x-name`` and
        ``iana-token`` values that they don't recognize without attempting
        to interpret or parse the value data.

    Returns:
        The VALUE parameter or the default.

    Examples:
        The VALUE defaults to the name of the property.
        Note that it is case-insensitive but always uppercase.

        .. code-block:: pycon

            >>> from icalendar import vBoolean
            >>> b = vBoolean(True)
            >>> b.VALUE
            'BOOLEAN'

        Setting the VALUE parameter of a typed property usually does not make sense.
        For convenience, using this property, the value will be converted to
        an uppercase string.
        If you have some custom property, you might use it like this:

        .. code-block:: pycon

            >>> from icalendar import vUnknown, Event
            >>> v = vUnknown("Some property text.")
            >>> v.VALUE = "x-type"  # lower case
            >>> v.VALUE
            'X-TYPE'
            >>> event = Event()
            >>> event.add("x-prop", v)
            >>> print(event.to_ical())
            BEGIN:VEVENT
            X-PROP;VALUE=X-TYPE:Some property text.
            END:VEVENT

    """
    value = self.params.value
    if value is None:
        _get_default_value = getattr(self, "_get_value", None)
        if _get_default_value is not None:
            value = _get_default_value()
    return self.default_value if value is None else value


def _set_value(self: VPROPERTY, value: str | None):
    """Set the VALUE parameter."""
    self.params.value = value


def _del_value(self: VPROPERTY):
    """Delete the VALUE parameter."""
    del self.params.value


VALUE = property(_get_value, _set_value, _del_value)

LABEL = string_parameter(
    "LABEL",
    """LABEL provides a human-readable label.

Conformance:
    This property parameter is specified in :rfc:`7986`,
    iCalendar Property Extensions.

    :rfc:`9253` makes use of this for the LINK property.
    This parameter maps to the "title"
    attribute defined in Section 3.4.1 of :rfc:`8288`.
    LABEL is used to label the destination
    of a link such that it can be used as a human-readable identifier
    (e.g., a menu entry) in the language indicated by the LANGUAGE
    (if present). The LABEL MUST NOT
    appear more than once in a given link; occurrences after the first
    MUST be ignored by parsers.

Description:
    This property parameter MAY be specified on the
    "CONFERENCE" property.  It is anticipated that other extensions to
    iCalendar will reuse this property parameter on new properties
    that they define.  As a result, clients MUST expect to find this
    property parameter present on many different properties.  It
    provides a human-readable label that can be presented to calendar
    users to allow them to discriminate between properties that might
    be similar or provide additional information for properties that
    are not self-describing.  The "LANGUAGE" property parameter can be
    used to specify the language of the text in the parameter value
    (as per Section 3.2.10 of :rfc:`5545`).

Examples:
    This is a label of a chat.

    .. code-block:: text

        CONFERENCE;VALUE=URI;FEATURE=VIDEO;
            LABEL="Web video chat, access code=76543";
            :https://video-chat.example.com/;group-id=1234

""",
)

FMTTYPE = string_parameter(
    "FMTTYPE",
    """FMTTYPE specfies the content type of a referenced object.

Conformance:
    :rfc:`5545` specifies the FMTTYPE.
    :rfc:`9253` adds FMTTYPE to LINK properties. In a LINK,
    FMTTYPE maps to the "type"
    attribute defined in Section 3.4.1 of :rfc:`8288`.
    See :rfc:`6838`.

Description:
    This parameter can be specified on properties that are
    used to reference an object.  The parameter specifies the media
    type :rfc:`4288` of the referenced object. For example, on the
    "ATTACH" property, an FTP type URI value does not, by itself,
    necessarily convey the type of content associated with the
    resource.  The parameter value MUST be the text for either an
    IANA-registered media type or a non-standard media type.

Example:
    A Microsoft Word document:

    .. code-block:: text

        ATTACH;FMTTYPE=application/msword:ftp://example.com/pub/docs/
         agenda.doc

    A website:

    .. code-block:: text

        LINK;FMTTYPE=text/html;LINKREL=SOURCE;LABEL=Venue;VALUE=URI:
         https://example.com/venue

    """,
)

LINKREL = string_parameter(
    "LINKREL",
    """LINKREL

Purpose:
    LINKREL specifies the relationship of data referenced
    by a LINK property.

Conformance:
    LINKREL is specified in :rfc:`9253`.
    This parameter maps to the link relation type defined in
    Section 2.1 of :rfc:`8288`.
    It is always quoted.

Description:
    This parameter MUST be specified on all LINK properties and define
    the type of reference.
    This allows programs consuming this data to automatically scan
    for references they support.
    There is no default relation type. Any link relation in the
    link registry established by :rfc:`8288`, or new link relations,
    may be used.
    It is expected that link relation types seeing significant usage
    in calendaring will have the calendaring usage described in an RFC.

    In the simplest case, a link relation type identifies the semantics
    of a link.  For example, a link with the relation type "copyright"
    indicates that the current link context has a copyright resource at
    the link target.

    Link relation types can also be used to indicate that the target
    resource has particular attributes, or exhibits particular
    behaviours; for example, a "service" link implies that the link
    target can be used as part of a defined protocol (in this case, a
    service description).

Registration:
    There are two kinds of relation types: registered and extension.
    These relation types are registered in :rfc:`8288`.

.. seealso::

    `Registered Link Relation Types
    <https://www.iana.org/assignments/link-relations/link-relations.xhtml>`_.


Examples:
    This identifies the latest version of the event information.

    .. code-block:: text

        LINKREL=latest-version

""",
)


def _get_GAP(prop) -> timedelta | None:  # noqa: N802
    """GAP

    Purpose:
        GAP specifies the length of the gap, positive or negative,
        between two components with a temporal relationship.

    Format Definition:
        Same as the DURATION value type defined in :rfc:`5545`, Section 3.3.6.

    Description:
        This parameter MAY be specified on the RELATED-TO property and defines
        the duration of time between the predecessor and successor in an interval.
        When positive, it defines the lag time between a task and its logical successor.
        When negative, it defines the lead time.

    Examples:
        An example of lag time might be if Task-A is "paint the room" and Task-B is
        "lay the carpets". Then, Task-A may be related to Task-B with
        RELTYPE=FINISHTOSTART with a gap of 1 day -- long enough for the paint to dry.

        .. code-block:: text

            ====================
            |  paint the room  |--+
            ====================  |
                                  |(lag of one day)
                                  |
                                  |  ===================
                                  +->| lay the carpet  |
                                     ===================

        For an example of lead time, in constructing a two-story building,
        the electrical work must be done before painting. However,
        the painter can move in to the first floor as the electricians move upstairs.

        .. code-block:: text

            =====================
            |  electrical work  |--+
            =====================  |
                     +-------------+
                     |(lead of estimated time)
                     |  ==================
                     +->|    painting    |
                        ==================
    """
    value = prop.params.get("GAP")
    if value is None:
        return None
    from icalendar.prop import vDuration

    if isinstance(value, str):
        return vDuration.from_ical(value)
    if not isinstance(value, vDuration):
        raise TypeError("Value MUST be a vDuration instance")
    return value.td


def _set_GAP(prop, value: timedelta | str | None):  # noqa: N802
    """Set the GAP parameter as a timedelta."""
    if value is None:
        prop.params.pop("GAP", None)
        return
    from icalendar.prop import vDuration

    prop.params["GAP"] = vDuration(value)


def _del_GAP(prop):  # noqa: N802
    """Delete the GAP parameter."""
    prop.params.pop("GAP", None)


GAP = property(_get_GAP, _set_GAP, _del_GAP)

__all__ = [
    "ALTREP",
    "CN",
    "CUTYPE",
    "DELEGATED_FROM",
    "DELEGATED_TO",
    "DIR",
    "FBTYPE",
    "FMTTYPE",
    "GAP",
    "LABEL",
    "LANGUAGE",
    "LINKREL",
    "MEMBER",
    "PARTSTAT",
    "RANGE",
    "RELATED",
    "ROLE",
    "RSVP",
    "SENT_BY",
    "TZID",
    "VALUE",
    "quoted_list_parameter",
    "string_parameter",
]
